# SPDX-License-Identifier: GPL-2.0+
import ctypes
import numpy as np

from .fVar_t import fVar_t

from .utils import copy_array


_all_dts = {}


def make_dt(name):
    class _fDerivedType(ctypes.Structure):
        pass

    _fDerivedType.__name__ = name
    return _fDerivedType


class fDT(fVar_t):
    def __init__(self, obj, fvar, allobjs=None, cvalue=None):
        self.obj = obj
        self.fvar = fvar
        self.allobjs = allobjs
        self.cvalue = cvalue
        self.unpack = True

        # Get obj for derived type spec
        self._dt_obj = self.allobjs[self.obj.dt_type()]
        self._dt_name = self._dt_obj.name

        self._init_args()

        self._ctype = None

    def _init_args(self):
        # Store fVar's for each component
        self._dt_args = {}
        for var in self._dt_obj.dt_components():
            # Catch dt's which contain themselves:
            if var.is_derived() and var.dt_type() == self.obj.dt_type():
                raise NotImplementedError(
                    "Derived types containing themselves not supported yet"
                )

            self._dt_args[var.name] = self.fvar(var, allobjs=self.allobjs)

    def ctype(self):
        if self._ctype is None:
            fields = []
            for var in self._dt_obj.dt_components():
                fields.append((var.name, self._dt_args[var.name].ctype()))

            class _fDerivedType(ctypes.Structure):
                _fields_ = fields

            if self._dt_name not in _all_dts:
                _all_dts[self._dt_name] = _fDerivedType
            else:
                try:
                    _all_dts[self._dt_name]._fields_ = fields
                except AttributeError:
                    pass

            self._ctype = _all_dts[self._dt_name]

        return self._ctype

    def from_ctype(self, ct):
        self.cvalue = ct
        return self.value

    def from_address(self, addr):
        self.cvalue = self.ctype().from_address(addr)
        return self.cvalue

    def in_dll(self, lib):
        self.cvalue = self.ctype().in_dll(lib, self.mangled_name)
        return self.cvalue

    def from_param(self, param):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        for key, value in param.items():
            if key not in self._dt_args:
                raise KeyError(f"{key} not present in {self._dt_obj.name}")

            if self._dt_args[key].obj.is_derived():
                self._dt_args[key] = self.fvar(
                    self.allobjs[key],
                    allobjs=self.allobjs,
                    cvalue=getattr(self.cvalue, key),
                )
                for k, v in value.items():
                    self._dt_args[key].__setitem__(k, v)
            else:
                if not isinstance(value, fVar_t):
                    self._dt_args[key].from_param(value)
                else:
                    self._dt_args[key].from_ctype(value.cvalue())

                v = self._dt_args[key].cvalue

                if self._dt_args[key].is_array:
                    copy_array(
                        ctypes.addressof(v),
                        ctypes.addressof(getattr(self.cvalue, key)),
                        1,
                        ctypes.sizeof(v),
                    )
                else:
                    setattr(self.cvalue, key, v)

        return self.cvalue

    @property
    def value(self):
        return self

    @value.setter
    def value(self, value):
        self.from_param(value)

    def keys(self):
        return [i.name for i in self._dt_obj.dt_components()]

    def values(self):
        return [self.__getitem__(key) for key in self.keys()]

    def items(self):
        return [(key, self.__getitem__(key)) for key in self.keys()]

    def __contains__(self, key):
        return key in self.keys()

    def __getitem__(self, key):
        if key in self._dt_args:
            return self._dt_args[key].from_ctype(getattr(self.cvalue, key))
        else:
            raise KeyError(f"{key} not present in {self._dt_obj.name}")

    def __setitem__(self, key, value):
        if key in self._dt_args:
            self.from_param({key: value})
        else:
            raise KeyError(f"{key} not present in {self._dt_obj.name}")

    def __dir__(self):
        return list(self._dt_args.keys())

    def __getattr__(self, key):
        if "_dt_args" in self.__dict__:
            if key in self._dt_args:
                raise AttributeError("Can't get components as attributes")

        if key in self.__dict__:
            return self.__dict__[key]
        else:
            raise AttributeError

    def __setattr__(self, key, value):
        if "_dt_args" in self.__dict__:
            if key in self._dt_args:
                raise AttributeError("Can't set components as attributes")

        if key == "value":
            self.from_param(value)

        self.__dict__[key] = value


class fExplicitDT(fVar_t):
    def __init__(self, obj, fvar, allobjs=None, cvalue=None):
        self.obj = obj
        self.fvar = fvar
        self.allobjs = allobjs
        self.cvalue = cvalue
        self.unpack = True

        # Get obj for derived type spec
        self._dt_obj = self.allobjs[self.obj.dt_type()]

        self._dt_ctype = fDT(self.obj, self.fvar, allobjs=self.allobjs)
        self._saved = {}

    def ctype(self):
        self._ctype = self._dt_ctype.ctype() * self.obj.size
        return self._ctype

    def __getitem__(self, index):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        if isinstance(index, tuple):
            ind = np.ravel_multi_index(index, self.obj.shape(), order="F")
        else:
            ind = index

        if ind > self.obj.size:
            raise IndexError("Out of bounds")

        if ind not in self._saved:
            self._saved[ind] = fDT(
                self.obj, self.fvar, allobjs=self.allobjs, cvalue=self.cvalue[ind]
            )

        return self._saved[ind]

    def __setitem__(self, index, value):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        if isinstance(index, tuple):
            ind = np.ravel_multi_index(index, self.obj.shape(), order="F")
        else:
            ind = index

        if ind > self.obj.size:
            raise IndexError("Out of bounds")

        if ind not in self._saved:
            self._saved[ind] = fDT(
                self.obj, allobjs=self.allobjs, cvalue=self.cvalue[ind]
            )

        self._saved[ind].value = value

    def from_param(self, param):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        for index, value in enumerate(param):
            self.__setitem__(index, value)

        return self.cvalue

    @property
    def value(self):
        return self

    @value.setter
    def value(self, value):
        self.from_param(value)



# SPDX-License-Identifier: GPL-2.0+

import ctypes
import numpy as np

from .fVar_t import fVar_t
from .fArrays import fAssumedShape
from .utils import copy_array, resolve_other_args


class fStr(fVar_t):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._len = None

    def ctype(self):
        return self._ctype_base * self.len()

    def from_param(self, value):
        if self.obj.is_deferred_len():
            self._len = len(value)

        if self.cvalue is None:
            self.cvalue = self.ctype()()

        self._value = value

        if hasattr(self._value, "encode"):
            self._value = self._value.encode()

        if len(self._value) > self.len():
            self._value = self._value[: self.len()]
        else:
            self._value = self._value + b" " * (self.len() - len(self._value))

        # self._buf = bytearray(self._value)  # Need to keep hold of the reference
        self.cvalue.value = self._value

        return self.cvalue

    @property
    def value(self):
        try:
            return self.cvalue.value.decode()
        except AttributeError:
            return str(self.cvalue)  # Functions returning str's give us str not bytes

    @value.setter
    def value(self, value):
        self.from_param(value)

    def len(self):
        if self._len is None:
            if self.obj.is_deferred_len():
                self._len = len(self.cvalue)
            else:
                self._len = self.obj.strlen.value
        return self._len

    def ctype_len(self, *args):
        return ctypes.c_int64(self.len())

    def __doc__(self):
        try:
            return f"{self.type}(LEN={self.obj.strlen}) :: {self.name}"
        except AttributeError:
            return f"{self.type}(LEN=:) :: {self.name}"

    def sizeof(self):
        return ctypes.sizeof(self.ctype)


class fAllocStr(fStr):
    def __init__(self, *args, **kwargs):
        self._len = None
        self._len_ctype = None
        super().__init__(*args, **kwargs)

    def ctype(self):
        return self._ctype_base

    @property
    def _ctype_base(self):
        return ctypes.c_char_p * self.len()

    @_ctype_base.setter
    def _ctype_base(self, value):
        return ctypes.c_char_p * self.len()

    def from_param(self, value):
        if value is None:
            return (ctypes.c_char_p * 1)()

        self._len = len(value)

        self._value = value

        if hasattr(self._value, "encode"):
            self._value = self._value.encode()

        if self.cvalue is None:
            self.cvalue = self.ctype()()
        elif len(value) != len(self.cvalue):
            self.cvalue = self.ctype().from_address(ctypes.addressof(self.cvalue))

        if len(self._value) > self.len():
            self._value = self._value[: self.len()]
        else:
            self._value = self._value + b" " * (self.len() - len(self._value))

        self.cvalue[0] = self._value

        return self.cvalue

    @property
    def value(self):
        try:
            x = self.cvalue[0]
        except Exception:
            return None

        if x is None:
            return None
        else:
            return x.decode()

    @value.setter
    def value(self, value):
        self.from_param(value)

    def len(self):
        if self._len_ctype is not None:
            self._len = self._len_ctype.contents.value

        if self._len is None:
            self._len = 1
        return self._len

    def in_dll(self, lib):
        self.cvalue = self.ctype().in_dll(lib, self.mangled_name)
        return self.cvalue

    def __doc__(self):
        return f"character(LEN=(:)), allocatable :: {self.name}"

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def to_proc(self, value, other_args):
        self.obj = resolve_other_args(self.obj, other_args)

        if value is None:
            l = 0
        else:
            l = len(value)

        self._len_ctype = ctypes.pointer(ctypes.c_int64(l))

        self.cvalue = self.from_param(value)

        return self.Args(None, self.cvalue, self._len_ctype)


class fStrExplicit(fStr):
    def __init__(self, *args, **kwargs):
        self._len_ctype = None
        self._len = None
        super().__init__(*args, **kwargs)
        self.unpack = False

    def ctype(self):
        return self._ctype_base

    @property
    def _ctype_base(self):
        return ctypes.c_char * self.len() * self.obj.size

    @_ctype_base.setter
    def _ctype_base(self, value):
        try:
            return ctypes.c_char * self.len() * self.obj.size
        except TypeError:
            pass  # Dont allways now size when loading class

    def _array_check(self, value, know_shape=True):
        shape = self.obj.shape()
        ndim = self.obj.ndim

        if not np.issubdtype(value.dtype, np.bytes_):
            raise TypeError("Character strings must be bytes (S dtype)")

        if not value.flags["F_CONTIGUOUS"]:
            value = np.asfortranarray(value)

        if value.ndim != ndim:
            raise ValueError(
                f"Wrong number of dimensions, got {value.ndim} expected {ndim}"
            )

        if know_shape:
            if not self.obj.is_allocatable and list(value.shape) != shape:
                raise ValueError(f"Wrong shape, got {value.shape} expected {shape}")

        value = value.ravel(order="F")
        return value

    def from_param(self, value):
        if value is None:
            raise ValueError("Character array must not be None")

        self._value = self._array_check(value)

        self._len = self._value.dtype.itemsize

        if self.cvalue is None:
            self.cvalue = self.ctype()()

        copy_array(
            self._value.ctypes.data,
            ctypes.addressof(self.cvalue),
            ctypes.sizeof(self._ctype_base),
            1,
        )

        return self.cvalue

    def from_ctype(self, ct):
        self.cvalue = ct

        return self.value

    @property
    def value(self):
        z = np.zeros(self.obj.shape(), dtype=f"S{self.len()}")

        copy_array(
            ctypes.addressof(self.cvalue),
            z.ctypes.data,
            ctypes.sizeof(self._ctype_base),
            1,
        )

        return z

    @value.setter
    def value(self, value):
        self.from_param(value)

    def len(self):
        if self._len_ctype is not None:
            self._len = self._len_ctype.value

        if self._len is None:
            if self.obj.is_deferred_len():
                self._len = len(self.cvalue)
            else:
                self._len = self.obj.strlen.value

        return self._len

    def in_dll(self, lib):
        self.cvalue = self.ctype().in_dll(lib, self.mangled_name)
        return self.cvalue

    def __doc__(self):
        return (
            f"character(LEN=({self.len()})),dimension({self.obj.shape}) :: {self.name}"
        )

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def to_proc(self, value, other_args):
        self.obj = resolve_other_args(self.obj, other_args)
        if value is None:
            l = 0
        else:
            l = value.dtype.itemsize

        self._len_ctype = ctypes.c_int64(l)

        self.cvalue = self.from_param(value)

        return self.Args(None, self.cvalue, self._len_ctype)

    def __del__(self):
        self.cvalue = None
        self._len_ctype = None


class fStrAssumedShape(fAssumedShape):
    def __init__(self, *args, **kwargs):
        self._len_ctype = None
        self._len = None
        super().__init__(*args, **kwargs)
        self.unpack = True
        self.is_array = True

    @property
    def _ctype_base(self):
        # print(self.len(),self.obj.size)
        return ctypes.c_char * int(self.len() * self.obj.size)

    @_ctype_base.setter
    def _ctype_base(self, value):
        try:
            return ctypes.c_char * int(self.len() * self.obj.size)
        except TypeError:
            pass  # Dont always now the size when loading the class

    def _array_check(self, value, know_shape=True):
        if not np.issubdtype(value.dtype, np.bytes_):
            raise TypeError("Character strings must be bytes (S dtype)")

        return super()._array_check(value, know_shape)

    @property
    def value(self):
        if hasattr(self.cvalue, "contents"):
            cv = self.cvalue.contents
        else:
            cv = self.cvalue

        if cv.base_addr is None:
            return None

        shape = []
        for i in range(self.obj.ndim):
            shape.append(cv.dims[i].ubound - cv.dims[i].lbound + 1)

        shape = tuple(shape)
        size = (np.prod(shape),)

        z = np.zeros(shape, dtype=f"S{self.len()}")

        copy_array(
            cv.base_addr,
            z.ctypes.data,
            ctypes.sizeof(ctypes.c_char * int(self.len())),
            np.prod(shape),
        )

        return z

    @value.setter
    def value(self, value):
        self.from_param(value)

    def len(self):
        if self._len_ctype is not None:
            self._len = self._len_ctype.value

        if self._len is None:
            if self.obj.is_deferred_len():
                self._len = len(self.cvalue)
            else:
                self._len = self.obj.strlen.value

        return self._len

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def to_proc(self, value, other_args):
        self.obj = resolve_other_args(self.obj, other_args)
        if value is None:
            l = self.obj.strlen.value
        else:
            l = value.dtype.itemsize

        self._len_ctype = ctypes.c_int64(l)

        self.cvalue = ctypes.pointer(self.from_param(value))

        return self.Args(None, self.cvalue, self._len_ctype)



# SPDX-License-Identifier: GPL-2.0+

import ctypes
import collections

from .utils import resolve_other_args


class fVar_t:
    Args = collections.namedtuple("arg", ["prepend", "arg", "append"])

    def __init__(self, obj, allobjs=None, cvalue=None):
        self.obj = obj
        self.allobjs = allobjs
        self.cvalue = cvalue
        self.unpack = True
        self.is_array = False

        self.type, self.kind = self.obj.type_kind()

        self._ctype_base = ctype_map(self.type, self.kind)

    @property
    def name(self):
        return self.obj.name

    @property
    def mangled_name(self):
        return self.obj.mangled_name

    @property
    def module(self):
        return self.obj.module

    def ctype_len(self):
        return None

    def from_ctype(self, ct):
        if hasattr(ct, "__ctypes_from_outparam__"):
            self.cvalue = ct
        else:  # Not actually a ctype
            self.from_param(ct)
            # Do it this way so we get the conversion code in value called (i.e decodeing bytes to a string)
        return self.value

    def from_address(self, addr):
        self.cvalue = self.ctype().from_address(addr)
        return self.cvalue

    def in_dll(self, lib):
        self.cvalue = self.ctype().in_dll(lib, self.mangled_name)
        return self.cvalue

    def to_proc(self, value, other_args):
        start = None
        arg = None
        end = None

        self.obj = resolve_other_args(self.obj, other_args)

        if self.obj.is_optional() and value is None:
            end = ctypes.c_byte(0)
            arg = None
            return self.Args(start, arg, end)

        raw_arg = self.from_param(value)
        if self.obj.is_optional():
            end = ctypes.c_byte(1)

        if self.obj.is_value():
            arg = raw_arg
        else:
            if self.obj.is_pointer():
                if self.obj.not_a_pointer():
                    arg = ctypes.pointer(raw_arg)
                else:
                    arg = ctypes.pointer(ctypes.pointer(raw_arg))
            else:
                arg = ctypes.pointer(raw_arg)

            if self.obj.is_deferred_len():
                end = self.ctype_len(value)

        return self.Args(start, arg, end)


def ctype_map(type, kind):
    if type == "INTEGER":
        if kind == 4:
            return ctypes.c_int32
        elif kind == 8:
            return ctypes.c_int64
        else:
            raise TypeError("Integer type of kind={kind} not supported")
    elif type == "REAL":
        if kind == 4:
            return ctypes.c_float
        elif kind == 8:
            return ctypes.c_double
        elif kind == 16:
            # Although we dont support quad yet we can keep things aligned
            return ctypes.c_ubyte * 16
        else:
            raise TypeError("Float type of kind={kind} not supported")
    elif type == "LOGICAL":
        return ctypes.c_int32
    elif type == "CHARACTER":
        return ctypes.c_char
    elif type == "COMPLEX":
        if kind == 4:
            ct = ctypes.c_float
        elif kind == 8:
            ct = ctypes.c_double
        elif kind == 16:
            ct = ctypes.c_ubyte * 16
        else:
            raise TypeError("Complex type of kind={kind} not supported")

        class complex(ctypes.Structure):
            _fields_ = [
                ("real", ct),
                ("imag", ct),
            ]

        return complex
    else:
        raise TypeError(f"Type={type} and kind={kind} not supported")



# SPDX-License-Identifier: GPL-2.0+

import subprocess
import pytest
import _pytest.skipping


def pytest_configure(config):
    subprocess.call(["make", "clean"], shell=True, cwd="tests")
    subprocess.call(["make"], shell=True, cwd="tests")


def pytest_sessionfinish(session, exitstatus):
    subprocess.call(["make", "clean"], shell=True, cwd="tests")



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/dt.{gf.lib_ext()}"
MOD = "./tests/dt.mod"

x = gf.fFort(SO, MOD)


# @pytest.mark.skip
class TestDTMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_dt_set_value(self):
        x.f_struct_simple["x"] = 1
        x.f_struct_simple["y"] = 0
        y = x.f_struct_simple
        self.assertEqual(y["x"], 1)
        self.assertEqual(y["y"], 0)

    def test_dt_set_dict(self):
        x.f_struct_simple = {"x": 5, "y": 5}
        y = x.f_struct_simple
        self.assertEqual(y["x"], 5)
        self.assertEqual(y["y"], 5)

    def test_dt_bad_dict(self):
        with pytest.raises(KeyError) as cm:
            x.f_struct_simple = {"asw": 2, "y": 0}

    def test_dt_bad_value(self):
        with pytest.raises(TypeError) as cm:
            x.f_struct_simple["x"] = "asde"

    def test_sub_dt_in_s_simple(self, capfd):
        y = x.sub_f_simple_in({"x": 1, "y": 10})
        out, err = capfd.readouterr()
        o = " ".join([str(i) for i in [1, 10]])
        self.assertEqual(out.strip(), o)

    def test_sub_dt_out_s_simple(self, capfd):
        y = x.sub_f_simple_out({})
        out, err = capfd.readouterr()
        self.assertEqual(y.args["x"]["x"], 1)
        self.assertEqual(y.args["x"]["y"], 10)

    def test_sub_dt_inout_s_simple(self, capfd):
        y = x.sub_f_simple_inout({"x": 5, "y": 3})
        out, err = capfd.readouterr()
        o = "  ".join([str(i) for i in [5, 3]])
        self.assertEqual(out.strip(), o)
        self.assertEqual(y.args["zzz"]["x"], 1)
        self.assertEqual(y.args["zzz"]["y"], 10)

    def test_sub_dt_inoutp_s_simple(self, capfd):
        y = x.sub_f_simple_inoutp({"x": 5, "y": 3})
        out, err = capfd.readouterr()
        o = "  ".join([str(i) for i in [5, 3]])
        self.assertEqual(out.strip(), o)
        self.assertEqual(y.args["zzz"]["x"], 1)
        self.assertEqual(y.args["zzz"]["y"], 10)

    def test_nested_dts(self):
        x.g_struct["a_int"] = 10
        self.assertEqual(x.g_struct["a_int"], 10)
        x.g_struct = {"a_int": 10, "f_struct": {"a_int": 3}}
        self.assertEqual(x.g_struct["f_struct"]["a_int"], 3)
        x.g_struct["f_struct"]["a_int"] = 8
        self.assertEqual(x.g_struct["f_struct"]["a_int"], 8)
        y = x.func_check_nested_dt()
        self.assertEqual(y.result, True)

    def test_func_set_f_struct(self):
        y = x.func_set_f_struct()
        self.assertEqual(y.result, True)

        self.assertEqual(x.f_struct["a_int"], 5)
        self.assertEqual(x.f_struct["a_int_lp"], 6)
        self.assertEqual(x.f_struct["a_real"], 7.0)
        self.assertEqual(x.f_struct["a_real_dp"], 8.0)
        self.assertEqual(x.f_struct["a_str"], "9999999999")

        v = np.array([9, 10, 11, 12, 13], dtype="int32")
        np.testing.assert_array_equal(x.f_struct["b_int_exp_1d"], v)

        v = np.array([9, 10, 11, 12, 13], dtype="int32")
        np.testing.assert_array_equal(x.e_int_target_1d, v)

    def test_func_set_f_struct_array_alloc(self):
        y = x.func_set_f_struct()

        v = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype="int32")
        np.testing.assert_array_equal(x.f_struct["c_int_alloc_1d"], v)

    def test_func_set_f_struct_array_ptr(self):
        y = x.func_set_f_struct()

        v = np.array([9, 10, 11, 12, 13], dtype="int32")
        np.testing.assert_array_equal(x.f_struct["d_int_point_1d"], v)

    # @pytest.mark.skip
    def test_recur_dt(self):  # Skip for now
        with pytest.raises(NotImplementedError) as cm:
            x.r_recur["a_int"] = 9
            self.assertEqual(x.r_recur["a_int"], 9)
            x.r_recur["s_recur"]["a_int"] = 9
            self.assertEqual(x.r_recur["s_recur"]["a_int"], 9)
            x.r_recur["s_recur"]["s_recur"]["a_int"] = 9
            self.assertEqual(x.r_recur["s_recur"]["s_recur"]["a_int"], 9)

    def test_arr_dt_exp_1d_set(self):
        x.g_struct_exp_1d[0]["a_int"] = 5
        self.assertEqual(x.g_struct_exp_1d[0]["a_int"], 5)
        x.g_struct_exp_1d[1]["a_int"] = 9
        self.assertEqual(x.g_struct_exp_1d[1]["a_int"], 9)
        self.assertEqual(
            x.g_struct_exp_1d[0]["a_int"], 5
        )  # recheck we didnt corrupt things

        with pytest.raises(IndexError) as cm:
            y = x.g_struct_exp_1d[10]

    def test_dt_not_array(self):
        with pytest.raises(KeyError) as cm:
            y = x.f_struct[9]["a_int"]

    def test_arr_dt_exp_2d_set(self):
        x.g_struct_exp_2d[0, 0]["a_int"] = 1
        x.g_struct_exp_2d[1, 0]["a_int"] = 2
        x.g_struct_exp_2d[0, 1]["a_int"] = 3
        x.g_struct_exp_2d[1, 1]["a_int"] = 4

        self.assertEqual(x.g_struct_exp_2d[0, 0]["a_int"], 1)
        self.assertEqual(x.g_struct_exp_2d[1, 0]["a_int"], 2)
        self.assertEqual(x.g_struct_exp_2d[0, 1]["a_int"], 3)
        self.assertEqual(x.g_struct_exp_2d[1, 1]["a_int"], 4)

        y = x.check_g_struct_exp_2d()
        self.assertEqual(y.result, True)

    def test_sub_struct_exp_1d(self):
        y = x.sub_struct_exp_1d({})

        s = y.args["x"]

        self.assertEqual(s[0]["a_int"], 5)
        self.assertEqual(s[1]["a_int"], 9)

        np.testing.assert_array_equal(
            s[0]["b_int_exp_1d"], np.array([66, 66, 66, 66, 66])
        )
        np.testing.assert_array_equal(
            s[1]["b_int_exp_1d"], np.array([77, 77, 77, 77, 77])
        )

    def test_fvar_as_arg(self, capfd):
        y = x.f_struct_simple
        y["x"] = 99
        y["y"] = 98

        z = x.sub_f_simple_in(y)
        out, err = capfd.readouterr()
        o = "99 98"
        self.assertEqual(out.strip(), o)

        z2 = x.sub_f_simple_in(z.args["x"])
        out, err = capfd.readouterr()
        o = "99 98"
        self.assertEqual(out.strip(), o)

    def test_func_return_s_struct_nested_2(self):
        y = x.func_return_s_struct_nested_2()
        self.assertEqual(y.result["a_int"], 123)
        self.assertEqual(y.result["f_nested"]["a_int"], 234)
        self.assertEqual(y.result["f_nested"]["f_struct"]["a_int"], 345)

    def test_derived_type_intent_out(self, capfd):
        # GH: #32
        p = {}
        y = x.derived_structure(p)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "10 20 30 40")

        assert np.all(y.args["p"]["iq"] == np.array([10, 20, 30, 40]))



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/proc_ptrs.{gf.lib_ext()}"
MOD = "./tests/proc_ptrs.mod"

x = gf.fFort(SO, MOD)


@pytest.mark.skip
class TestProcPtrsMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_proc_ptr_ffunc(self):
        x.sub_null_proc_ptr()
        with pytest.raises(AttributeError) as cm:
            y = x.p_func_func_run_ptr(1)

        x.p_func_func_run_ptr = x.func_func_run
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 10)
        y = x.p_func_func_run_ptr(2)
        self.assertEqual(y.result, 20)

        y = x.func_proc_ptr(5)
        y2 = x.p_func_func_run_ptr(5)
        self.assertEqual(y.result, y2.result)

    def test_proc_ptr_ffunc2(self):
        x.sub_null_proc_ptr()
        with pytest.raises(AttributeError) as cm:
            y = x.p_func_func_run_ptr2(1)  # Allready set

        x.p_func_func_run_ptr2 = x.func_func_run
        y = x.p_func_func_run_ptr2(10)
        self.assertEqual(y.result, 100)

    def test_proc_update(self):
        x.sub_null_proc_ptr()
        x.p_func_func_run_ptr = x.func_func_run
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 10)

        x.p_func_func_run_ptr = x.func_func_run2
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 2)

    def test_proc_func_arg(self):
        y = x.func_func_arg_dp(5, x.func_real)
        self.assertEqual(y.result, 500)

        y = x.func_func_arg(x.func_func_run)
        self.assertEqual(y.result, 10)

    def test_proc_proc_func_arg(self):
        x.sub_null_proc_ptr()
        x.p_func_func_run_ptr = x.func_func_run

        y = x.proc_proc_func_arg(x.p_func_func_run_ptr)
        self.assertEqual(y.result, 90)



# SPDX-License-Identifier: GPL-2.0+
import ctypes
import numpy as np

from .fVar_t import fVar_t

from .fScalars import fScalar, fCmplx
from .fArrays import fExplicitArr, fAssumedShape, fAssumedSize
from .fStrings import fStr, fAllocStr, fStrExplicit, fStrAssumedShape
from .fDT import fDT, fExplicitDT
from .fProcPtr import fProcPointer


class fVar:
    def __new__(cls, obj, *args, **kwargs):
        if obj.is_derived():
            if obj.is_array():
                if obj.is_explicit():
                    return fExplicitDT(obj, fVar, *args, **kwargs)
                raise NotImplementedError
            else:
                return fDT(obj, fVar, *args, **kwargs)
        elif obj.is_proc_pointer():
            raise NotImplementedError
            # return fProcPointer(obj, *args, **kwargs)
        elif obj.is_array():
            if obj.is_char():
                if obj.is_explicit():
                    return fStrExplicit(obj, *args, **kwargs)
                elif obj.is_assumed_shape() or obj.is_allocatable() or obj.is_pointer():
                    return fStrAssumedShape(obj, *args, **kwargs)
                else:
                    raise TypeError(
                        "Unknown type of character array not currently supported"
                    )
            elif obj.is_explicit():
                return fExplicitArr(obj, *args, **kwargs)
            elif obj.is_assumed_size():
                return fAssumedSize(obj, *args, **kwargs)
            elif obj.is_assumed_shape() or obj.is_allocatable() or obj.is_pointer():
                return fAssumedShape(obj, *args, **kwargs)
            else:
                raise TypeError("Unknown array type")
        else:
            if obj.is_char():
                if obj.is_allocatable():
                    return fAllocStr(obj, *args, **kwargs)
                else:
                    return fStr(obj, *args, **kwargs)
            elif obj.is_complex():
                return fCmplx(obj, *args, **kwargs)
            else:
                return fScalar(obj, *args, **kwargs)



# SPDX-License-Identifier: GPL-2.0+

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata  # type: ignore

try:
    __version__ = metadata.version("gfort2py")
except metadata.PackageNotFoundError:
    # package is not installed
    pass



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/dummy_arrays.{gf.lib_ext()}"
MOD = "./tests/dummy_arrays.mod"

x = gf.fFort(SO, MOD)


# @pytest.mark.skip
class TestDummyArrayMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_sub_alloc_1d_arrs(self):
        y = x.sub_alloc_int_1d_arrs()

    def test_c_int_alloc_1d_non_alloc(self):
        y = x.sub_alloc_int_1d_cleanup()
        self.assertEqual(x.c_int_alloc_1d, None)

    def test_ndarray(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()

        self.assertEqual(np.sum(x.c_int_alloc_1d), 5)

    def test_c_int_alloc_1d(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_int_alloc_1d, v)

    def test_c_int_alloc_2d(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_int_alloc_2d, v)

    def test_c_int_alloc_3d(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_int_alloc_3d, v)

    def test_c_int_alloc_4d(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_int_alloc_4d, v)

    def test_c_int_alloc_5d(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_int_alloc_5d, v)

    def test_c_int_alloc_1d_set(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5])
        v[:] = 5
        x.c_int_alloc_1d = v
        np.testing.assert_array_equal(x.c_int_alloc_1d, v)
        self.assertEqual(np.sum(x.c_int_alloc_1d), 25)

    def test_c_int_alloc_2d_set(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 5
        x.c_int_alloc_2d = v
        np.testing.assert_array_equal(x.c_int_alloc_2d, v)

    def test_c_int_alloc_3d_set(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 5
        x.c_int_alloc_3d = v
        np.testing.assert_array_equal(x.c_int_alloc_3d, v)

    def test_c_int_alloc_4d_set(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 5
        x.c_int_alloc_4d = v
        np.testing.assert_array_equal(x.c_int_alloc_4d, v)

    @pytest.mark.skip
    def test_c_int_alloc_5d_set(self):
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 5
        x.c_int_alloc_5d = v
        np.testing.assert_array_equal(x.c_int_alloc_5d, v)

    @pytest.mark.skip
    def test_c_int_alloc_1d_large(self):
        # Can have issues exiting when using large (>255) arrays
        y = x.sub_alloc_int_1d_cleanup()
        y = x.sub_alloc_int_1d_arrs()
        v = np.zeros([256], dtype="int32")
        v[:] = 5
        x.c_int_alloc_1d = v
        np.testing.assert_array_equal(x.c_int_alloc_1d, v)

    def test_c_real_alloc_1d(self):
        y = x.sub_alloc_real_1d_cleanup()
        y = x.sub_alloc_real_1d_arrs()
        v = np.zeros([5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_alloc_1d, v)

    def test_c_real_alloc_2d(self):
        y = x.sub_alloc_real_1d_cleanup()
        y = x.sub_alloc_real_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_alloc_2d, v)

    def test_c_real_alloc_3d(self):
        y = x.sub_alloc_real_1d_cleanup()
        y = x.sub_alloc_real_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_alloc_3d, v)

    def test_c_real_alloc_4d(self):
        y = x.sub_alloc_real_1d_cleanup()
        y = x.sub_alloc_real_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_alloc_4d, v)

    def test_c_real_alloc_5d(self):
        y = x.sub_alloc_real_1d_cleanup()
        y = x.sub_alloc_real_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_alloc_5d, v)

    def test_c_real_dp_alloc_1d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5])
        v[:] = 2.0
        x.c_real_dp_alloc_1d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_1d, v)

    def test_c_real_dp_alloc_2d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_2d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_2d, v)

    def test_c_real_dp_alloc_3d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_3d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_3d, v)

    def test_c_real_dp_alloc_4d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_4d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_4d, v)

    @pytest.mark.skip
    def test_c_real_dp_alloc_5d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_5d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_5d, v)

    def test_c_real_dp_alloc_1d(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_dp_alloc_1d, v)

    def test_c_real_dp_alloc_2d(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_dp_alloc_2d, v)

    def test_c_real_dp_alloc_3d(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_dp_alloc_3d, v)

    def test_c_real_dp_alloc_4d(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_dp_alloc_4d, v)

    def test_c_real_dp_alloc_5d(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 1
        np.testing.assert_array_equal(x.c_real_dp_alloc_5d, v)

    def test_c_real_dp_alloc_1d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5])
        v[:] = 2.0
        x.c_real_dp_alloc_1d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_1d, v)

    def test_c_real_dp_alloc_2d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_2d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_2d, v)

    def test_c_real_dp_alloc_3d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_3d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_3d, v)

    @pytest.mark.skip
    def test_c_real_dp_alloc_4d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_4d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_4d, v)

    @pytest.mark.skip
    def test_c_real_dp_alloc_5d_set(self):
        y = x.sub_alloc_real_dp_1d_cleanup()
        y = x.sub_alloc_real_dp_1d_arrs()
        v = np.zeros([5, 5, 5, 5, 5])
        v[:] = 2.0
        x.c_real_dp_alloc_5d = v
        np.testing.assert_array_equal(x.c_real_dp_alloc_5d, v)

    def test_func_assumed_shape_arr_1d(self):
        v = np.zeros([5], dtype="int32")
        v[0] = 2.0
        y = x.func_assumed_shape_arr_1d(v)
        self.assertEqual(y.result, True)
        np.testing.assert_array_equal(y.args["x"], np.array([9, 9, 9, 9, 9]))

    def test_func_assumed_shape_arr_2d(self):
        v = np.zeros([5, 5], dtype="int32")
        v[1, 0] = 2.0
        y = x.func_assumed_shape_arr_2d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_shape_arr_3d(self):
        v = np.zeros([5, 5, 5], dtype="int32")
        v[2, 1, 0] = 2.0
        y = x.func_assumed_shape_arr_3d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_shape_arr_4d(self):
        v = np.zeros([5, 5, 5, 5], dtype="int32")
        v[3, 2, 1, 0] = 2.0
        y = x.func_assumed_shape_arr_4d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_shape_arr_5d(self):
        v = np.zeros([5, 5, 5, 5, 5], dtype="int32")
        v[4, 3, 2, 1, 0] = 2.0
        y = x.func_assumed_shape_arr_5d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_size_arr_1d(self):
        v = np.zeros([5], dtype="int32")
        v[1] = 2
        y = x.func_assumed_size_arr_1d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_size_arr_real_1d(self):
        v = np.zeros([5], dtype="float32")
        v[1] = 2.0
        y = x.func_assumed_size_arr_real_1d(v)
        self.assertEqual(y.result, True)

    def test_func_assumed_size_arr_real_dp_1d(self):
        v = np.zeros([5], dtype="float64")
        v[1] = 2.0
        y = x.func_assumed_size_arr_real_dp_1d(v)
        self.assertEqual(y.result, True)

    def test_sub_alloc_arr_1d(self):
        y = x.sub_alloc_arr_1d(None)
        vTest = np.zeros(10)
        vTest[:] = 10
        np.testing.assert_array_equal(y.args["x"], vTest)

    def test_logical_arr(self):
        xarr = np.zeros(10)
        x2arr = np.zeros(10)
        x2arr[:] = False
        xarr[:] = True

        y = x.func_alltrue_arr_1d(xarr)
        y2 = x.func_allfalse_arr_1d(x2arr)
        self.assertEqual(y.result, True)
        self.assertEqual(y2.result, True)

    def test_sub_arr_assumed_rank_int_1d(self, capfd):
        v = np.arange(10, 15)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_arr_assumed_rank_int_1d(v)
        out, err = capfd.readouterr()
        np.testing.assert_array_equal(y.args["zzz"], np.array([100] * 5))

    def test_sub_arr_assumed_rank_real_1d(self, capfd):
        v = np.arange(10.0, 15.0)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_arr_assumed_rank_real_1d(v)
        out, err = capfd.readouterr()
        np.testing.assert_array_equal(y.args["zzz"], np.array([100.0] * 5))

    def test_sub_arr_assumed_rank_dp_1d(self, capfd):
        v = np.arange(10.0, 15.0)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_arr_assumed_rank_dp_1d(v)
        out, err = capfd.readouterr()
        np.testing.assert_array_equal(y.args["zzz"], np.array([100.0] * 5))

    @pytest.mark.skip
    def test_sub_check_alloc_int_2d(self):
        arr_test = np.zeros((3, 4), dtype=np.int32, order="F")
        y = x.sub_check_alloc_int_2d(None)
        y2 = x.sub_check_alloc_int_2d(arr_test)
        np.testing.assert_array_equal(y.args["x"], y2.args["x"])

        z = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])

        np.testing.assert_array_equal(y.args["x"], z)

    @pytest.mark.skip
    def test_sub_check_alloc_int_3d(self):
        arr_test = np.zeros((3, 4, 5), dtype=np.int32, order="F")
        y = x.sub_check_alloc_int_3d(None)
        y2 = x.sub_check_alloc_int_3d(arr_test)
        np.testing.assert_array_equal(y.args["x"], y2.args["x"])

        z = np.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [5, 6, 7, 8, 9],
                    [9, 10, 11, 12, 13],
                    [13, 14, 15, 16, 17],
                ],
                [
                    [5, 6, 7, 8, 9],
                    [9, 10, 11, 12, 13],
                    [13, 14, 15, 16, 17],
                    [17, 18, 19, 20, 21],
                ],
                [
                    [9, 10, 11, 12, 13],
                    [13, 14, 15, 16, 17],
                    [17, 18, 19, 20, 21],
                    [21, 22, 23, 24, 25],
                ],
            ]
        )

        np.testing.assert_array_equal(y.args["x"], z)



# SPDX-License-Identifier: GPL-2.0+
import ctypes
import numpy as np

from .fVar_t import fVar_t
from .utils import copy_array


_index_t = ctypes.c_int64
_size_t = ctypes.c_int64


class _bounds14(ctypes.Structure):
    _fields_ = [("stride", _index_t), ("lbound", _index_t), ("ubound", _index_t)]


class _dtype_type(ctypes.Structure):
    _fields_ = [
        ("elem_len", _size_t),
        ("version", ctypes.c_int32),
        ("rank", ctypes.c_byte),
        ("type", ctypes.c_byte),
        ("attribute", ctypes.c_ushort),
    ]


def _make_fAlloc15(ndims):
    class _fAllocArray(ctypes.Structure):
        _fields_ = [
            ("base_addr", ctypes.c_void_p),
            ("offset", _size_t),
            ("dtype", _dtype_type),
            ("span", _index_t),
            ("dims", _bounds14 * ndims),
        ]

    return _fAllocArray


class fArray_t(fVar_t):
    def _array_check(self, value, know_shape=True):
        value = value.astype(self.obj.dtype())
        shape = self.obj.shape()
        ndim = self.obj.ndim

        if not value.flags["F_CONTIGUOUS"]:
            value = np.asfortranarray(value)

        if value.ndim != ndim:
            raise ValueError(
                f"Wrong number of dimensions, got {value.ndim} expected {ndim}"
            )

        if know_shape:
            if not self.obj.is_allocatable and list(value.shape) != shape:
                raise ValueError(f"Wrong shape, got {value.shape} expected {shape}")

        value = value.ravel(order="F")
        return value

    @property
    def ndim(self):
        return self.obj.ndim


class fExplicitArr(fArray_t):
    def ctype(self):
        return self._ctype_base * self.obj.size

    def from_param(self, value):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        self._value = self._array_check(value)
        copy_array(
            self._value.ctypes.data,
            ctypes.addressof(self.cvalue),
            ctypes.sizeof(self._ctype_base),
            self.obj.size,
        )
        return self.cvalue

    @property
    def value(self):
        return np.ctypeslib.as_array(self.cvalue).reshape(self.obj.shape(), order="F")

    @value.setter
    def value(self, value):
        self.from_param(value)

    def __doc__(self):
        return f"{self.type}(KIND={self.kind})({self.obj.shape()}) :: {self.name}"

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def len(self):
        return len(self._value)


class fAssumedShape(fArray_t):
    _BT_UNKNOWN = 0
    _BT_INTEGER = _BT_UNKNOWN + 1
    _BT_LOGICAL = _BT_INTEGER + 1
    _BT_REAL = _BT_LOGICAL + 1
    _BT_COMPLEX = _BT_REAL + 1
    _BT_DERIVED = _BT_COMPLEX + 1
    _BT_CHARACTER = _BT_DERIVED + 1
    _BT_CLASS = _BT_CHARACTER + 1
    _BT_PROCEDURE = _BT_CLASS + 1
    _BT_HOLLERITH = _BT_PROCEDURE + 1
    _BT_VOID = _BT_HOLLERITH + 1
    _BT_ASSUMED = _BT_VOID + 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_array = True

    def ctype(self):
        return _make_fAlloc15(self.obj.ndim)

    def from_param(self, value):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        if value is not None:
            self._value = self._array_check(value, False)

            # copy_array(
            #     self._value.ctypes.data,
            #     self.cvalue.base_addr,
            #     ctypes.sizeof(self._ctype_base()),
            #     np.size(value)
            # )
            self.cvalue.base_addr = self._value.ctypes.data

            self.cvalue.span = ctypes.sizeof(self._ctype_base())

            strides = []
            shape = np.shape(value)
            for i in range(self.ndim):
                self.cvalue.dims[i].lbound = _index_t(1)
                self.cvalue.dims[i].ubound = _index_t(shape[i])
                strides.append(
                    self.cvalue.dims[i].ubound - self.cvalue.dims[i].lbound + 1
                )

            spans = []
            for i in range(self.ndim):
                spans.append(int(np.prod(strides[:i])))
                self.cvalue.dims[i].stride = _index_t(spans[-1])

            self.cvalue.offset = -np.sum(spans)

        self.cvalue.dtype.elem_len = self.cvalue.span
        self.cvalue.dtype.version = 0
        self.cvalue.dtype.rank = self.ndim
        self.cvalue.dtype.type = self.ftype()
        self.cvalue.dtype.attribute = 0

        return self.cvalue

    @property
    def value(self):
        if self.cvalue.base_addr is None:
            return None

        shape = []
        for i in range(self.obj.ndim):
            shape.append(self.cvalue.dims[i].ubound - self.cvalue.dims[i].lbound + 1)

        shape = tuple(shape)
        size = (np.prod(shape),)

        PTR = ctypes.POINTER(self._ctype_base)
        x_ptr = ctypes.cast(self.cvalue.base_addr, PTR)

        return np.ctypeslib.as_array(x_ptr, shape=size).reshape(shape, order="F")

    @value.setter
    def value(self, value):
        self.from_param(value)

    def __doc__(self):
        return f"{self.type}(KIND={self.kind})(:) :: {self.name}"

    def ftype(self):
        if self.obj.type() == "INTEGER":
            return self._BT_INTEGER
        elif self.obj.type() == "LOGICAL":
            return self._BT_LOGICAL
        elif self.obj.type() == "REAL":
            return self._BT_REAL
        elif self.obj.type() == "COMPLEX":
            return self._BT_COMPLEX
        elif self.obj.type() == "CHARACTER":
            return self._BT_CHARACTER

        raise NotImplementedError(
            f"Assumed shape array of type {self.type} and kind {self.kind} not supported yet"
        )

    def __del__(self):
        if self.cvalue is not None:
            self.cvalue.base_addr = None

    def print(self):
        if self.cvalue is None:
            return ""

        print(f"base_addr {self.cvalue.base_addr}")
        print(f"offset {self.cvalue.offset}")
        print(f"dtype")
        print(f"\t elem_len {self.cvalue.dtype.elem_len}")
        print(f"\t version {self.cvalue.dtype.version}")
        print(f"\t rank {self.cvalue.dtype.rank}")
        print(f"\t type {self.cvalue.dtype.type}")
        print(f"\t attribute {self.cvalue.dtype.attribute}")
        print(f"span {self.cvalue.span}")
        print(f"dims {self.ndim}")
        for i in range(self.ndim):
            print(f"\t lbound {self.cvalue.dims[i].lbound}")
            print(f"\t ubound {self.cvalue.dims[i].ubound}")
            print(f"\t stride {self.cvalue.dims[i].stride}")


class fAssumedSize(fArray_t):
    def ctype(self):
        return self._ctype_base * np.prod(self._value.shape)

    def from_param(self, value):
        self._value = self._array_check(value)
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        copy_array(
            self._value.ctypes.data,
            ctypes.addressof(self.cvalue),
            ctypes.sizeof(self._ctype_base),
            np.size(value),
        )
        return self.cvalue

    @property
    def value(self):
        return np.ctypeslib.as_array(self.cvalue, shape=np.size(self._value)).reshape(
            self._value.shape, order="F"
        )

    @value.setter
    def value(self, value):
        self.from_param(value)

    def __doc__(self):
        return f"{self.type}(KIND={self.kind})(*) :: {self.name}"

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def len(self):
        return len(self._value)

    def ctype_len(self, *args):
        return ctypes.c_int64(self.len())

    @property
    def ndim(self):
        return 1



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/basic.{gf.lib_ext()}"
MOD = "./tests/basic.mod"

x = gf.fFort(SO, MOD)


class TestBasicMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_mising_var(self):
        with pytest.raises(AttributeError) as cm:
            a = x.invalid_var

    def test_a_int(self):
        v = 1
        x.a_int = v
        self.assertEqual(x.a_int, v)

    def test_a_int_str(self):
        with pytest.raises(TypeError) as cm:
            x.a_int = "abc"

    def test_a_real(self):
        v = 1.0
        x.a_real = v
        self.assertEqual(x.a_real, v)

    def test_a_real_str(self):
        with pytest.raises(TypeError) as cm:
            x.a_real = "abc"

    def test_const_int_set(self):
        with pytest.raises(AttributeError) as cm:
            x.const_int = 2

    def test_const_int(self):
        self.assertEqual(x.const_int, 1)

    def test_const_int_p1(self):
        self.assertEqual(x.const_int_p1, 2)

    def test_const_int_long(self):
        self.assertEqual(x.const_int_lp, 1)

    def test_const_real_dp(self):
        self.assertEqual(x.const_real_dp, 1.0)

    def test_const_real_pi_dp(self):
        self.assertEqual(x.const_real_pi_dp, 3.14)

    def test_const_real_qp(self):
        self.assertEqual(x.const_real_qp, 1.0)

    def test_sub_no_args(self, capfd):
        x.sub_no_args()
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "1")

    def test_sub_alter_mod(self):
        y = x.sub_alter_mod()
        self.assertEqual(x.a_int, 99)
        self.assertEqual(x.a_int_lp, 99)
        self.assertEqual(x.a_real, 99.0)
        self.assertEqual(x.a_real_dp, 99.0)

    @pytest.mark.skip("Skipping due to quad support")
    def test_sub_alter_mod_qp(self):
        y = x.sub_alter_mod()
        self.assertEqual(x.a_real_qp, 99.0)

    def test_func_int_in(self):
        v = 5
        y = x.func_int_in(v)
        self.assertEqual(int(y.result), 2 * v)

    def test_func_int_in_multi(self):
        v = 5
        w = 3
        u = 4
        y = x.func_int_in_multi(v, w, u)
        self.assertEqual(y.result, v + w + u)

    def test_sub_int_in(self, capfd):
        v = 5

        y = x.sub_int_in(v)
        out, err = capfd.readouterr()
        self.assertEqual(int(out), 2 * v)

    def test_func_int_no_args(self):
        y = x.func_int_no_args()
        self.assertEqual(y.result, 2)

    def test_func_real_no_args(self):
        y = x.func_real_no_args()
        self.assertEqual(y.result, 3.0)

    def test_func_real_dp_no_args(self):
        y = x.func_real_dp_no_args()
        self.assertEqual(y.result, 4.0)

    def test_sub_int_out(self, capfd):
        v = 5

        y = x.sub_int_out(v)
        out, err = capfd.readouterr()
        self.assertEqual(y.args, {"x": 1})

    def test_sub_int_inout(self, capfd):
        v = 5

        y = x.sub_int_inout(v)
        out, err = capfd.readouterr()
        self.assertEqual(y.args, {"x": 2 * v})

    def test_sub_int_no_intent(self, capfd):
        v = 5

        y = x.sub_int_no_intent(v)
        out, err = capfd.readouterr()
        self.assertEqual(y.args, {"x": 2 * v})

    def test_sub_real_inout(self, capfd):
        v = 5.0

        y = x.sub_real_inout(v)
        out, err = capfd.readouterr()
        self.assertEqual(y.args, {"x": 2 * v})

    def test_func_return_res(self):
        y = x.func_return_res(2)
        self.assertEqual(y.result, True)
        y = x.func_return_res(10)
        self.assertEqual(y.result, False)

    def test_sub_int_p(self, capfd):
        y = x.sub_int_p(1)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "1")
        self.assertEqual(y.args["zzz"], 5)

    def test_sub_real_p(self, capfd):
        y = x.sub_real_p(1.0)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "1.00")
        self.assertEqual(y.args["zzz"], 5.0)

    def test_sub_opt(self, capfd):
        y = x.sub_int_opt(1)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "100")

        y = x.sub_int_opt(None)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "200")

    def test_sub_opt_val(self, capfd):
        y = x.sub_int_opt_val(1)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "100")

        y = x.sub_int_opt_val(None)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), "200")

    def test_second_mod(self):
        y = x.sub_use_mod()
        self.assertEqual(x.test2_x, 1)

    def test_func_value(self):
        y = x.func_int_value(5)
        self.assertEqual(y.result, 10)

    def test_func_pass_mod_var(self):
        x.a_int = 5
        z = x.func_int_in(x.a_int)
        self.assertEqual(z.result, 10)

    def test_sub_man_args(self):
        # if this doesn't seg fault we are good
        x.sub_many_args(
            1, 2, 3, 4, True, False, True, "abc", "def", "ghj", "qwerty", "zxcvb"
        )

    def test_func_intent_out(self):
        y = x.func_intent_out(9, 0)
        self.assertEqual(y.result, 9)
        self.assertEqual(y.args["x"], 9)

        y = x.func_intent_out(x=0, y=9)  # Swap order of arguments
        self.assertEqual(y.result, 9)
        self.assertEqual(y.args["x"], 9)

    def test_func_result(self):
        y = x.func_result(9, 0)
        self.assertEqual(y.result, 18)
        self.assertEqual(y.args["y"], 9)

    def test_func_bool(self):
        y = x.func_test_bool(1)
        assert type(y.result) == bool
        self.assertEqual(y.result, True)
        y = x.func_test_bool(0)
        assert type(y.result) == bool
        self.assertEqual(y.result, False)

    def test_negatives(self):
        assert x.const_neg_int == -1
        assert x.const_neg_real == -3.14

    def test_logical_parammeters(self):
        assert x.const_logical_true
        assert not x.const_logical_false



# SPDX-License-Identifier: GPL-2.0+

import ctypes

from .fVar_t import fVar_t


class fProcPointer(fVar_t):
    def ctype(self):
        return self._func

    def from_param(self, param):
        # memmove?
        raise NotImplementedError

    @property
    def value(self):
        return self

    @value.setter
    def value(self, value):
        self.from_param(value)



# SPDX-License-Identifier: GPL-2.0+

import ctypes
import itertools

from .fUnary import run_unary


def copy_array(src, dst, length, size):
    ctypes.memmove(
        dst,
        src,
        length * size,
    )


def resolve_other_args(obj, other_args):
    """
    We want to iterate over the components of obj
    and if they are symbol_refs look them up in other_args.

    This way we can resolve attributes that are runtime set,
    for instance dimension(n) arrays, where n is a dummy argument

    For now limit to array bounds
    """

    if not obj.is_array():
        return obj

    for i in itertools.chain(obj.sym.array_spec.lower, obj.sym.array_spec.upper):
        i = _resolve_arg(i, other_args)

    return obj


def _resolve_arg(arg, other_args):
    if arg.exp_type == "CONSTANT":
        return arg
    elif arg.exp_type == "VARIABLE":
        ref = arg.value.ref
        for j in other_args:
            if ref == j.symbol_ref:
                arg.value = j.value
    elif arg.exp_type == "OP":
        # Unary operator
        op = arg.unary_op
        arg1 = _resolve_arg(arg.unary_args[0], other_args)
        arg2 = _resolve_arg(arg.unary_args[1], other_args)

        # print(op,arg1.value,arg2.value)
        # print(run_unary(op,arg1.value,arg2.value))
        arg.value = run_unary(op, arg1.value, arg2.value)

    return arg



# SPDX-License-Identifier: GPL-2.0+

import ctypes
import numpy as np

from .fVar_t import fVar_t


class fScalar(fVar_t):
    def ctype(self):
        return self._ctype_base

    def from_param(self, param):
        if self.cvalue is None:
            self.cvalue = self.ctype()(param)
        else:
            self.cvalue.value = param
        return self.cvalue

    @property
    def value(self):
        if self.type == "INTEGER":
            return int(self.cvalue.value)
        elif self.type == "REAL":
            if self.kind == 16:
                raise NotImplementedError(f"Quad precision floats not supported yet")
            elif self.kind == 8:
                return np.double(self.cvalue.value)
            else:
                return float(self.cvalue.value)
        elif self.type == "LOGICAL":
            return self.cvalue.value == 1

    @value.setter
    def value(self, value):
        self.from_param(value)

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def __doc__(self):
        return f"{self.type}(KIND={self.kind}) :: {self.name}"


class fCmplx(fVar_t):
    def ctype(self):
        return self._ctype_base

    def from_param(self, param):
        if self.cvalue is None:
            self.cvalue = self.ctype()()

        self.cvalue.real = param.real
        self.cvalue.imag = param.imag
        return self.cvalue

    @property
    def value(self):
        x = self.cvalue

        if self.kind == 16:
            raise NotImplementedError(
                f"Quad precision complex numbers not supported yet"
            )
        return complex(x.real, x.imag)

    @value.setter
    def value(self, value):
        self.from_param(value)

    def sizeof(self):
        return ctypes.sizeof(self.ctype)

    def __doc__(self):
        return f"{self.type}(KIND={self.kind}) :: {self.name}"



# SPDX-License-Identifier: GPL-2.0+

from pyparsing import OneOrMore, nestedExpr
from dataclasses import dataclass
import numpy as np
import gzip
import sys

import pprint

import typing as t


def string_clean(string):
    if string is None:
        return
    if string.startswith("'") or string.startswith('"'):
        string = string[1:]
    if string.endswith("'") or string.endswith('"'):
        string = string[:-1]

    return string


class VersionError(Exception):
    pass


class NotAnArrayError(Exception):
    pass


#################################


@dataclass
class s_item:
    name: str
    ambiguous: bool
    id: int

    def __post_init__(self):
        self.name = string_clean(self.name)
        self.ambiguous = self.ambiguous != "0"
        self.id = int(self.id)


class Summary:
    def __init__(self, summ):
        self._item_id = {}
        self._item_name = {}

        for i in range(0, len(summ), 3):
            d = s_item(*summ[i : i + 3])
            self._item_id[d.id] = d
            self._item_name[d.name] = d

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._item_id[key]
        if isinstance(key, str):
            return self._item_name[key]
        else:
            raise TypeError(f"Dont understand type of {key}")

    def keys(self):
        return list(self._item_id.keys()) + list(self._item_name.keys())

    def __contains__(self, key):
        if isinstance(key, int):
            return key in self._item_id
        if isinstance(key, str):
            return key in self._item_name
        else:
            raise TypeError(f"Dont understand type of {key}")

    def names(self):
        return self._item_name.keys()


#################################


@dataclass
class c_item:
    name: str
    id: int
    saved_flag: bool
    _unknown: int
    _unknown2: str

    def __post_init__(self):
        self.name = string_clean(self.name)
        self.id = int(self.id)
        self.saved_flag = int(self.saved_flag)
        self._unknown = int(self._unknown)
        self._unknown2 = string_clean(self._unknown2)


#################################


@dataclass(init=False)
class generics:
    name: str = ""
    module: str = ""
    id: t.List[int] = -1

    def __init__(self, *args):
        self.name = string_clean(args[0])
        self.module = string_clean(args[1])
        self.id = []
        for i in args[2:]:
            self.id.append(int(i))


#################################


def hextofloat(s, double=False):
    # Given hex like parameter '0.12decde@9' returns 5065465344.0
    man, exp = s.split("@")
    exp = int(exp)
    decimal = man.index(".")
    negative = man[0] == "-"
    man = man[decimal + 1 :]
    man = man.ljust(exp, "0")
    man = man[:exp] + "." + man[exp:]
    man = man + "P0"
    if negative:
        man = "-" + man
    if double:
        return np.double.fromhex(man)
    else:
        return float.fromhex(man)


#####################################


def print_args(x):
    print()
    for i in x:
        print(i)
    print()


class utils:
    def shape(self):
        if self.is_array():
            return self.sym.array_spec.pyshape
        else:
            raise NotAnArrayError("Not an array")

    def dtype(self):
        t = self.type()
        k = int(self.kind())

        if t == "INTEGER" or t == "LOGICAL":
            if k == 4:
                return "i4"
            elif k == 8:
                return "i8"
        elif t == "REAL":
            if k == 4:
                return "f4"
            elif k == 8:
                return "f8"
        elif t == "COMPLEX":
            if k == 4:
                return "c4"
            elif k == 8:
                return "c8"
        elif t == "CHARACTER":
            return f"S{self.strlen.value}"

        raise NotImplementedError(f"Object of type {t} and kind {k} not supported yet")

    def type(self):
        return self.sym.ts.type

    def ref(self):
        return self.head.id

    def flavor(self):
        return self.sym.attr.flavor

    def kind(self):
        return self.sym.ts.kind

    def is_pointer(self):
        return "POINTER" in self.sym.attr.attributes

    def is_parameter(self):
        return self.flavor() == "PARAMETER"

    def is_value(self):
        return "VALUE" in self.sym.attr.attributes

    def is_optional(self):
        return "OPTIONAL" in self.sym.attr.attributes

    def is_optional_value(self):
        return self.is_optional() and self.is_value()

    def is_char(self):
        return self.type() == "CHARACTER"

    def is_variable(self):
        return self.flavor() == "VARIABLE"

    def is_procedure(self):
        return self.flavor() == "PROCEDURE"

    def is_proc_pointer(self):
        return "PROC_POINTER" in self.sym.attr.attributes

    def is_logical(self):
        return self.sym.ts.type == "LOGICAL"

    def is_complex(self):
        return self.sym.ts.type == "COMPLEX"

    def is_subroutine(self):
        return self.sym.sym_ref.ref == 0

    def is_function(self):
        return self.sym.sym_ref.ref != 0

    def is_array(self):
        return "DIMENSION" in self.sym.attr.attributes

    def is_always_explicit(self):
        return "ALWAYS_EXPLICIT" in self.sym.attr.attributes

    def is_dummy(self):
        return "DUMMY" in self.sym.attr.attributes

    def is_allocatable(self):
        return "ALLOCATABLE" in self.sym.attr.attributes

    def needs_array_desc(self):
        return self.is_dummy() or self.is_allocatable() or self.is_always_explicit()

    def not_a_pointer(self):
        return (
            self.needs_array_desc()
            and self.is_array()
            and not self.is_assumed_shape()
            and not self.is_assumed_size()
        )

    def is_explicit(self):
        return (
            self.sym.array_spec.array_type == "EXPLICIT"
            and not self.is_always_explicit()
        )

    def is_assumed_size(self):
        return self.sym.array_spec.array_type == "ASSUMED_SIZE"

    def is_assumed_shape(self):
        return self.sym.array_spec.array_type == "ASSUMED_SHAPE"

    def is_deferred_len(self):
        # Only needed for things that need an extra function argument for their length
        if self.is_char():
            try:
                return self.sym.ts.charlen.value == -1
            except AttributeError:
                return True
        elif self.is_array():
            return self.is_assumed_size()

        return False

    def is_derived(self):
        return self.sym.ts.type == "DERIVED"

    def is_pdt_def(self):
        return "PDT_TEMPLATE" in self.sym.attr.attributes

    @property
    def strlen(self):
        if self.is_char() and not self.is_deferred_len():
            return self.sym.ts.charlen
        raise AttributeError("Not a deferred length type")

    @property
    def ndim(self):
        if self.is_array():
            return self.sym.array_spec.rank
        else:
            raise NotAnArrayError("Not an array")

    @property
    def size(self):
        if self.is_array():
            return np.prod(self.shape())
        else:
            raise NotAnArrayError("Not an array")

    def value(self):
        if self.is_parameter():
            v = self.sym.parameter.value
            if not self.is_array():
                return v
            else:
                return np.array(v, dtype=self.dtype()).reshape(self.shape(), order="F")
        else:
            raise AttributeError("Not a parameter")

    def type_kind(self):
        return self.type(), self.kind()

    def return_arg(self):
        if self.is_procedure():
            return self.sym.sym_ref.ref
        raise AttributeError("Not a procedure")

    def args(self):
        if self.is_procedure():
            return self.sym.formal_arg
        raise AttributeError("Not a procedure")

    def dt_type(self):
        return self.sym.ts.class_ref.ref

    def dt_components(self):
        return self.sym.comp


@dataclass(init=False)
class attribute:
    flavor: str = ""
    intent: str = ""
    proc: str = ""
    if_source: str = ""
    save: str = ""
    ext_attr: int = -1
    extension: int = -1
    attributes: t.Set[str] = None

    def __init__(self, *args):
        self.flavor = string_clean(args[0])
        self.intent = string_clean(args[1])
        self.proc = string_clean(args[2])
        self.if_source = string_clean(args[3])
        self.save = string_clean(args[4])
        self.ext_attr = int(args[5])
        self.extension = int(args[6])
        self.attributes = set([string_clean(i) for i in args[7:]])


@dataclass
class namespace:
    ref: int = -1

    def __post_init__(self):
        self.ref = symbol_ref(self.ref)


@dataclass
class header:
    id: int
    name: str  # If first letter is capitalized then its a dt
    module: str
    bindc: str
    parent_id: int

    def __post_init__(self):
        self.id = int(self.id)
        self.parent_id = int(self.parent_id)
        self.name = string_clean(self.name)
        self.module = string_clean(self.module)
        self.bindc = len(string_clean(self.bindc)) > 0

    @property
    def mn_name(self):
        if self.module:
            return f"__{self.module}_MOD_{self.name}"


@dataclass
class symbol_ref:
    ref: int = -1

    def __post_init__(self):
        self.ref = int(self.ref)


@dataclass(init=False)
class formal_arglist:
    symbol: t.List[symbol_ref] = None

    def __init__(self, *args):
        self.symbol = []
        for i in args:
            self.symbol.append(symbol_ref(i))

    def __len__(self):
        return len(self.symbol)

    def __iter__(self):
        return iter(self.symbol)


@dataclass(init=False)
class typebound_proc:
    name: str = ""
    access: str = ""
    overridable: str = ""
    nopass: str = ""
    is_generic: str = ""
    ppc: str = ""
    pass_arg: str = ""
    pass_arg_num: symbol_ref = None
    proc_ref: symbol_ref = None

    def __init__(self, *args, **kwargs):
        self.name = string_clean(args[0][0])
        self.access = args[0][1][0]
        self.overridable = args[0][1][1]
        self.nopass = args[0][1][2]
        self.is_generic = args[0][1][3]
        self.ppc = args[0][1][4]
        self.pass_arg = string_clean(args[0][1][5])
        self.pass_arg_num = symbol_ref(args[0][1][6])

        # TODO: Handle is_generic
        self.proc_ref = symbol_ref(args[0][1][7][0])


@dataclass(init=False)
class derived_ns:
    unknown1: str = None
    proc: t.List[typebound_proc] = None

    def __init__(self, *args, **kwargs):
        if not len(args):
            return
        self.unknown1 = args[0]
        self.proc = []
        for i in args[1]:
            self.proc.append(typebound_proc(i))


@dataclass(init=False)
class actual_arglist:
    args: None
    kwargs: None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


@dataclass(init=False)
class typespec:
    type: str = ""
    kind: int = -1  # If class symbol_ref else kind
    class_ref: symbol_ref = None  # If class/derived type symbol_ref else kind
    interface: symbol_ref = None
    is_c_interop: int = -1
    is_iso_c: int = -1
    type2: str = ""  # Repeat of type
    charlen: int = -1  # If character
    deferred_cl: bool = False  # if character and deferred length
    args = None

    def __init__(self, *args):
        self.args = args
        self.type = args[0]
        if self.type == "CLASS" or self.type == "DERIVED":
            self.class_ref = symbol_ref(args[1])
        else:
            self.kind = int(args[1])

        if len(args[2]):
            self.interface = symbol_ref(args[2])

        self.is_c_interop = bool(int(args[3]))
        self.is_iso_c = bool(int(args[4]))
        self.type2 = args[5]
        try:
            if not args[6][0]:
                self.charlen = -1
            else:
                self.charlen = expression(
                    *args[6][0]
                )  # TODO: might this need to be iterated for mulit-d strings?
        except IndexError:
            self.charlen = -1

        try:
            self.deferred_cl = args[7] == "DEFERRED_CL"
        except (TypeError, IndexError):
            self.deferred_cl = False


@dataclass(init=False)
class expression:
    exp_type: str = ""
    ts: typespec = None
    rank: int = -1
    _value: t.Any = None
    _resolved_value: t.Any = None  # value may by a symbol_ref, so this is the value after resolving the reference
    arglist: actual_arglist = None  # PDT's?
    charlen: int = -1
    unary_op: str = ""
    unary_args: t.Any = None
    _unknown: t.Any = None

    def __init__(self, *args):
        self._resolved_value = None
        if not len(args):
            return
        self.exp_type = args[0]
        self.ts = typespec(*args[1])
        self.rank = int(args[2])

        if self.exp_type == "OP":
            self._value = None
            self.unary_op = args[3]
            self.unary_args = [expression(*args[4]), expression(*args[5])]
            try:
                self._unknown = args[6]  # What is this for?
            except IndexError:
                pass
        elif self.exp_type == "FUNCTION":
            self._value = symbol_ref(args[3])
        elif self.exp_type == "CONSTANT":
            if self.ts.type == "REAL":
                self._value = hextofloat(string_clean(args[3]), self.ts.kind == 8)
            elif self.ts.type == "INTEGER":
                self._value = int(string_clean(args[3]))
            elif self.ts.type == "CHARACTER":
                self.charlen = int(args[3])
                self._value = string_clean(args[4])
            elif self.ts.type == "COMPLEX":
                self._value = complex(
                    hextofloat(string_clean(args[3]), self.ts.kind == 8),
                    hextofloat(string_clean(args[4]), self.ts.kind == 8),
                )
            elif self.ts.type == "LOGICAL":
                self._value = int(args[3]) == 1
            else:
                raise NotImplementedError(args)
        elif self.exp_type == "VARIABLE":
            self._value = symbol_ref(args[3])
        elif self.exp_type == "SUBSTRING":
            raise NotImplementedError(args)
        elif self.exp_type == "ARRAY":
            self._value = []
            for i in args[3]:
                self._value.append(
                    expression(*i[0]).value
                )  # Wheres the extra component comming from?
        elif self.exp_type == "NULL":
            self._value = args[3]
        elif self.exp_type == "COMPCALL":
            raise NotImplementedError(args)
        elif self.exp_type == "PPC":
            raise NotImplementedError(args)
        else:
            raise AttributeError(f"Can't match {self.exp_type}")

    @property
    def value(self):
        if self._resolved_value is not None:
            return self._resolved_value
        else:
            return self._value

    @value.setter
    def value(self, value):
        self._resolved_value = value


@dataclass(init=False)
class arrayspec:
    rank: int = -1
    corank: int = -1
    array_type: str = ""
    lower: t.List[expression] = None
    upper: t.List[expression] = None

    def __init__(self, *args):
        if not len(args):
            return

        self.rank = int(args[0])
        self.corank = int(args[1])
        self.array_type = args[2]
        self.lower = []
        self.upper = []
        for i in range(self.rank + self.corank):
            if len(args[3 + i * 2]):
                self.lower.append(expression(*args[3 + i * 2]))
            if len(args[4 + i * 2]):
                self.upper.append(expression(*args[4 + i * 2]))

    @property
    def fshape(self):
        res = []
        for l, u in zip(self.lower, self.upper):
            res.append([l.value, u.value])

        return res

    @property
    def pyshape(self):
        res = []
        if self.lower is None:
            return []

        for l, u in zip(self.lower, self.upper):
            res.append(u.value - l.value + 1)

        return res

    @property
    def size(self):
        return np.prod(self.pyshape)


@dataclass(init=False)
class component(utils):
    id: int = -1
    name: str = ""
    ts: typespec = None
    array_spec: arrayspec = None
    expr: expression = None
    actual_arg: actual_arglist = None
    attr: attribute = None
    access: str = ""
    initializer: expression = None
    proc_ptr: typebound_proc = None

    def __init__(self, *args):
        args = list(args)

        self.id = int(args[0])
        self.name = string_clean(args[1])
        self.ts = typespec(*args[2])
        self.array_spec = arrayspec(*args[3])
        if len(args[4]):
            self.expr = expression(*args[4])
        if len(args[5]):
            self.actual_arg = actual_arglist(*args[5])
        self.attr = attribute(*args[6])
        self.access = string_clean(args[7])

        if self.name == "_final" or self.name == "_hash":
            self.initializer = expression(*args[8])
            _ = args.pop(8)

        if not self.attr.proc == "UNKNOWN-PROC":
            self.proc_ptr = typebound_proc(args[8])

        # This lets us reuse the code for accessing symbols
        # inside the parent utils class
        self.sym = self


@dataclass(init=False)
class components:
    comp: t.List[component] = None

    def __init__(self, *args):
        self.comp = []
        for i in args:
            self.comp.append(component(*i))

    def __len__(self):
        return len(self.comp)

    def __iter__(self):
        return iter(self.comp)


@dataclass(init=False)
class namelist:
    sym_ref: t.List[symbol_ref] = None

    def __init__(self, *args):
        self.sym_ref = []
        if len(args):
            for i in args:
                self.sym_ref.append(symbol_ref(i))


@dataclass(init=False)
class simd_dec:
    args: None
    kwargs: None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


@dataclass(init=False)
class data:
    attr: attribute
    comp: components = None
    comp_access: str = ""  # Only for DT's
    ts: typespec = None
    ns: namespace = None
    common_link: symbol_ref = None
    formal_arg: formal_arglist = None
    parameter: expression = None  # If parameter
    array_spec: arrayspec = None
    sym_ref: symbol_ref = None
    sym_ref_cray: symbol_ref = None  # If cray_pointer
    derived: derived_ns = None
    actual_arg: actual_arglist = None
    nml: namelist = None
    intrinsic: int = -1
    intrinsic_symbol: int = -1
    hash: int = -1
    simd: simd_dec = None

    def __init__(self, *args):
        args = list(
            args
        )  # Do it this was as there are optional terms we may need to pop
        self.attr = attribute(*args[0])
        self.comp = components(*args[1])

        if isinstance(args[2], str):
            self.comp_access = args[2]
            _ = args.pop(2)

        self.ts = typespec(*args[2])
        self.ns = namespace(args[3])
        self.common_link = symbol_ref(args[4])
        self.formal_arg = formal_arglist(*args[5])
        if self.attr.flavor == "PARAMETER":
            self.parameter = expression(*args[6])
            _ = args.pop(6)
        self.array_spec = arrayspec(*args[6])
        if True:
            self.sym_ref = symbol_ref(args[7])
        else:
            pass  # Ignore cray pointers
        self.derived = derived_ns(*args[8])
        self.actual_arg = actual_arglist(*args[9])
        self.nml = namelist(*args[10])
        self.intrinsic = int(args[11])
        if len(args) > 12:
            self.intrinsic_symbol = int(args[12])
        if len(args) > 13:
            self.hash = int(args[13])
        if len(args) > 14:
            if args[15] is not None:
                self.simd = simd_dec(*args[14])


@dataclass(init=False)
class symbol(utils):
    head: header = None
    sym: data = None
    raw: str = ""

    def __init__(self, *args):
        self.head = header(*args[0:5])
        self.sym = data(*args[5])
        self.raw = args

    @property
    def name(self):
        return self.head.name

    @property
    def mangled_name(self):
        return self.head.mn_name


class module(object):
    version = 15

    def __init__(self, filename, load_only=False):
        self.filename = filename

        with gzip.open(self.filename) as f:
            x = f.read().decode()

        self.mod_info = x[: x.index("\n")]

        v = int(self.mod_info.split("'")[1])

        if v != self.version:
            raise VersionError("Unsupported module version")

        data = x[x.index("\n") + 1 :].replace("\n", " ")

        self.parsed_data = OneOrMore(nestedExpr()).parseString(data)

        if not load_only:
            self.interface = self.parsed_data[0]

            self.operators = self.parsed_data[1]
            self.generics = self.proc_generics(self.parsed_data[2])

            self.common = self.proc_common(self.parsed_data[3])
            self.equivalence = self.parsed_data[4]

            self.omp = self.parsed_data[5]

            self.symbols = self.parse_symbols(self.parsed_data[6])
            self.summary = Summary(self.parsed_data[7])

    def parse_symbols(self, data):
        result = {}
        for i in range(0, len(data), 6):
            s = symbol(*data[i : i + 6])
            result[s.head.id] = s

        return result

    def proc_common(self, data):
        result = {}
        for i in data:
            d = c_item(*i)
            result[d.id] = d
        return result

    def proc_generics(self, data):
        result = {}
        for i in data:
            d = generics(*i)
            result[d.name] = d
        return result

    def keys(self):
        return self.summary.names()

    def __contains__(self, key):
        return key in self.keys()

    def __getitem__(self, key):
        try:
            return self.symbols[self.summary[key].id]
        except KeyError:
            # Not a global variable maybe a function argument?
            return self.symbols[key]


if __name__ == "__main__":
    m = module(filename=sys.argv[1])
    for i in m.keys():
        pprint.pprint(m[i])



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/complex.{gf.lib_ext()}"
MOD = "./tests/comp.mod"

x = gf.fFort(SO, MOD)


class TestComplexMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_a_const_cmplx(self):
        self.assertEqual(x.const_cmplx, complex(1.0, 1.0))

    def test_a_const_cmplx_dp(self):
        self.assertEqual(x.const_cmplx_dp, complex(1.0, 1.0))

    def test_a_const_cmplx_qp(self):
        self.assertEqual(x.const_cmplx_qp, complex(1.0, 1.0))

    def test_a_cmplx(self):
        v = complex(1.0, 1.0)
        x.a_cmplx = v
        self.assertEqual(x.a_cmplx, v)

    def test_a_cmplx_dp(self):
        v = complex(1.0, 1.0)
        x.a_cmplx_dp = v
        self.assertEqual(x.a_cmplx_dp, v)

    def test_a_cmplx_qp(self):
        v = complex(1.0, 1.0)
        with pytest.raises(TypeError) as cm:
            x.a_cmplx_qp = v
            self.assertEqual(x.a_cmplx_qp, v)

    def test_sub_cmplx_inout(self):
        v = complex(1.0, 1.0)
        y = x.sub_cmplx_inout(v)
        self.assertEqual(y.args["c"], v * 5)

    def test_func_cmplx_value(self):
        v = complex(1.0, 1.0)
        y = x.sub_cmplx_value(v, v)
        self.assertEqual(y.args["cc"], v * 5)

    def test_func_ret_cmplx(self):
        v = complex(1.0, 1.0)
        y = x.func_ret_cmplx(v)
        self.assertEqual(y.result, v * 5)



# SPDX-License-Identifier: GPL-2.0+
import ctypes
import numpy as np
import os
import platform

from .module_parse import module

from .fVar import fVar
from .fProc import fProc
from .fParameters import fParam

_TEST_FLAG = os.environ.get("_GFORT2PY_TEST_FLAG") is not None


class fFort:
    _initialized = False

    def __init__(self, libname, mod_file):
        self._lib = ctypes.CDLL(libname)
        self._mod_file = mod_file
        self._module = module(self._mod_file)

        self._saved = {}
        self._initialized = True

    def keys(self):
        return self._module.keys()

    def __contains__(self, key):
        return key in self._module.keys()

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]

        if "_initialized" in self.__dict__:
            if self._initialized:
                if key not in self.keys():
                    raise AttributeError(f"{self._mod_file}  has no attribute {key}")

            if self._module[key].is_variable():
                if key not in self._saved:
                    self._saved[key] = fVar(self._module[key], allobjs=self._module)
                self._saved[key].in_dll(self._lib)
                return self._saved[key].value
            elif self._module[key].is_proc_pointer():
                # Must come before fProc
                if key not in self._saved:
                    self._saved[key] = fVar(self._module[key], allobjs=self._module)
                return self._saved[key]
            elif self._module[key].is_procedure():
                return fProc(self._lib, self._module[key], self._module)
            elif self._module[key].is_parameter():
                return fParam(self._module[key]).value
            else:
                raise NotImplementedError(
                    f"Object type {self._module[key].flavor()} not implemented yet"
                )

    def __setattr__(self, key, value):
        if key in self.__dict__:
            self.__dict__[key] = value
            return

        if "_initialized" in self.__dict__:
            if self._initialized:
                if self._module[key].is_variable():
                    if key not in self._saved:
                        self._saved[key] = fVar(self._module[key], allobjs=self._module)
                    self._saved[key].in_dll(self._lib)
                    self._saved[key].value = value
                    return
                elif self._module[key].is_parameter():
                    raise AttributeError("Can not alter a parameter")
                elif self._module[key].is_proc_pointer():
                    if key not in self._saved:
                        self._saved[key] = fVar(self._module[key], allobjs=self._module)
                    self._saved[key].value = value
                    return
                else:
                    raise NotImplementedError(
                        f"Object type {self._module[key].flavor()} not implemented yet"
                    )

        self.__dict__[key] = value

    @property
    def __doc__(self):
        return f"MODULE={self._module.filename}"

    def __str__(self):
        return f"{self._module.filename}"


def mod_info(mod_file):
    return module(mod_file)


def lib_ext():
    os = platform.system()
    if os == "Darwin":
        return "dylib"
    elif os == "Windows":
        return "dll"
    else:
        return "so"



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/.{gf.lib_ext()}"
MOD = "./tests/.mod"

# x=gf.fFort(SO,MOD)


class TestStringMethods:
    pass



# SPDX-License-Identifier: GPL-2.0+


class fParam:
    def __init__(self, obj):
        self.obj = obj

    @property
    def value(self):
        return self.obj.value()

    @value.setter
    def value(self, value):
        raise AttributeError("Parameters can't be altered")

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)

    @property
    def module(self):
        return self._value.module



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/pdt.{gf.lib_ext()}"
MOD = "./tests/pdt.mod"

x = gf.fFort(SO, MOD)


@pytest.mark.skip
class TestOOMethods:
    def test_p_proc_call(self):
        pass



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/explicit_arrays.{gf.lib_ext()}"
MOD = "./tests/explicit_arrays.mod"

x = gf.fFort(SO, MOD)


class TestExplicitArrayMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_const_int_arr_error(self):
        with pytest.raises(AttributeError) as cm:
            x.const_int_arr = "abc"

    def test_const_int_arr(self):
        np.testing.assert_array_equal(
            x.const_int_arr, np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0], dtype="int32")
        )

    def test_const_real_arr(self):
        np.testing.assert_array_equal(
            x.const_real_arr,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 0.0], dtype="float"),
        )

    def test_const_dp_arr(self):
        np.testing.assert_array_equal(
            x.const_real_dp_arr,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 0.0], dtype="float"),
        )

    def test_b_int_exp_1d(self):
        v = np.random.randint(0, 100, size=(5))
        x.b_int_exp_1d = v
        np.testing.assert_array_equal(x.b_int_exp_1d, v)

    def test_b_int_exp_2d(self):
        v = np.asfortranarray(np.random.randint(0, 100, size=(5, 5)), dtype="int32")
        x.b_int_exp_2d = v
        np.testing.assert_array_equal(x.b_int_exp_2d, v)

    def test_b_int_exp_3d(self):
        v = np.random.randint(0, 100, size=(5, 5, 5))
        x.b_int_exp_3d = v
        np.testing.assert_array_equal(x.b_int_exp_3d, v)

    def test_b_int_exp_4d(self):
        v = np.random.randint(0, 100, size=(5, 5, 5, 5))
        x.b_int_exp_4d = v
        np.testing.assert_array_equal(x.b_int_exp_4d, v)

    def test_b_int_exp_5d(self):
        v = np.random.randint(0, 100, size=(5, 5, 5, 5, 5))
        x.b_int_exp_5d = v
        np.testing.assert_array_equal(x.b_int_exp_5d, v)

    def test_b_real_exp_1d(self):
        v = np.random.random(size=(5))
        x.b_real_exp_1d = v
        np.testing.assert_allclose(x.b_real_exp_1d, v)

    def test_b_real_exp_2d(self):
        v = np.random.random(size=(5, 5))
        x.b_real_exp_2d = v
        np.testing.assert_allclose(x.b_real_exp_2d, v)

    def test_b_real_exp_3d(self):
        v = np.random.random(size=(5, 5, 5))
        x.b_real_exp_3d = v
        np.testing.assert_allclose(x.b_real_exp_3d, v)

    def test_b_real_exp_4d(self):
        v = np.random.random(size=(5, 5, 5, 5))
        x.b_real_exp_4d = v
        np.testing.assert_allclose(x.b_real_exp_4d, v)

    def test_b_real_exp_5d(self):
        v = np.random.random(size=(5, 5, 5, 5, 5))
        x.b_real_exp_5d = v
        np.testing.assert_allclose(x.b_real_exp_5d, v)

    def test_b_real_exp_5d_2(self):
        v = np.random.random(size=(2, 3, 4, 5, 6))
        x.b_real_exp_5d_2 = v
        np.testing.assert_allclose(x.b_real_exp_5d_2, v)

    def test_b_real_dp_exp_1d(self):
        v = np.random.random(size=(5))
        x.b_real_dp_exp_1d = v
        np.testing.assert_allclose(x.b_real_dp_exp_1d, v)

    def test_b_real_dp_exp_2d(self):
        v = np.random.random(size=(5, 5))
        x.b_real_dp_exp_2d = v
        np.testing.assert_allclose(x.b_real_dp_exp_2d, v)

    def test_b_real_dp_exp_3d(self):
        v = np.random.random(size=(5, 5, 5))
        x.b_real_dp_exp_3d = v
        np.testing.assert_allclose(x.b_real_dp_exp_3d, v)

    def test_b_real_dp_exp_4d(self):
        v = np.random.random(size=(5, 5, 5, 5))
        x.b_real_dp_exp_4d = v
        np.testing.assert_allclose(x.b_real_dp_exp_4d, v)

    def test_b_real_dp_exp_5d(self):
        v = np.random.random(size=(5, 5, 5, 5, 5))
        x.b_real_dp_exp_5d = v
        np.testing.assert_allclose(x.b_real_dp_exp_5d, v)

    @pytest.mark.skip
    def test_sub_array_n_int_1d(self, capfd):
        v = np.arange(0, 5)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_array_n_int_1d(np.size(v), v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    @pytest.mark.skip
    def test_sub_array_n_int_2d(self, capfd):
        v = [0, 1, 2, 3, 4] * 5
        v = np.array(v).reshape(5, 5)
        o = " ".join([str(i) for i in np.asfortranarray(v).flatten(order="F")])

        y = x.sub_array_n_int_2d(5, 5, v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_int_1d(self, capfd):
        v = np.arange(0, 5)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_exp_array_int_1d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_int_2d(self, capfd):
        v = np.arange(0, 5 * 5).reshape((5, 5))
        o = "".join(
            [str(i).zfill(2).ljust(3) for i in np.asfortranarray(v).flatten(order="F")]
        )
        y = x.sub_exp_array_int_2d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_int_3d(self, capfd):
        v = np.arange(0, 5 * 5 * 5).reshape((5, 5, 5))
        o = "".join(
            [str(i).zfill(3).ljust(4) for i in np.asfortranarray(v).flatten(order="F")]
        )
        y = x.sub_exp_array_int_3d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_real_1d(self, capfd):
        v = np.arange(0, 5.0).reshape((5))
        o = "  ".join(
            ["{:>4.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )
        y = x.sub_exp_array_real_1d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_real_2d(self, capfd):
        v = np.arange(0, 5.0 * 5.0).reshape((5, 5))
        o = "  ".join(
            ["{:>4.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )
        y = x.sub_exp_array_real_2d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_real_3d(self, capfd):
        v = np.arange(0, 5.0 * 5.0 * 5.0).reshape((5, 5, 5))
        o = " ".join(
            ["{:>5.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )
        y = x.sub_exp_array_real_3d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_int_1d_multi(self, capfd):
        u = 19
        w = 20
        v = np.arange(0, 5)
        o = " ".join([str(i) for i in np.asfortranarray(v).flatten(order="F")])
        y = x.sub_exp_array_int_1d_multi(u, v, w)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), str(u) + " " + o.strip() + " " + str(w))

    def test_sub_exp_array_real_dp_1d(self, capfd):
        v = np.arange(0, 5.0).reshape((5))
        o = "  ".join(
            ["{:>4.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )

        y = x.sub_exp_array_real_dp_1d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_real_dp_2d(self, capfd):
        v = np.arange(0, 5.0 * 5.0).reshape((5, 5))
        o = "  ".join(
            ["{:>4.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )

        y = x.sub_exp_array_real_dp_2d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_array_real_dp_3d(self, capfd):
        v = np.arange(0, 5.0 * 5.0 * 5.0).reshape((5, 5, 5))
        o = " ".join(
            ["{:>5.1f}".format(i) for i in np.asfortranarray(v).flatten(order="F")]
        )

        y = x.sub_exp_array_real_dp_3d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_sub_exp_inout(self, capfd):
        v = np.array([1, 2, 3, 4, 5])

        y = x.sub_exp_inout(v)
        out, err = capfd.readouterr()

        np.testing.assert_array_equal(y.args["x"], 2 * v)

    def test_sub_arr_exp_p(self, capfd):
        v = np.arange(0, 5)
        o = " ".join([str(i) for i in v.flatten(order="F")])

        y = x.sub_exp_array_int_1d(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), o.strip())

    def test_logical_arr_multi(self):
        xarr = np.zeros(5)
        xarr[:] = True

        y = x.func_logical_multi(1.0, 2.0, xarr, 3.0, 4.0)
        self.assertEqual(y.result, True)

    @pytest.mark.skip
    def test_mesh_exp(self):
        # Github issue #13
        i = 5
        y = x.func_mesh_exp(i)
        np.all(y.result == np.arrange(0, i + 1 + 1))

    def test_mesh_exp2(self):
        i = 5
        z = np.zeros(i + 1)
        y = x.func_mesh_exp2(z, i)
        assert np.all(y.args["x"] == np.arange(1, i + 1 + 1))

    def test_mesh_exp3(self):
        i = 5
        z = np.zeros((i * 2) + 1)
        y = x.func_mesh_exp3(z, i)
        assert np.all(y.args["x"] == np.arange(1, (i * 2) + 1 + 1))

    def test_check_exp_2d_2m3(self):
        # Github issue #19
        arr_test = np.zeros((3, 4), dtype=int, order="F")

        arr_test[0, 1] = 1
        arr_test[1, 0] = 2
        arr_test[1, 2] = 3
        arr_test[-2, -1] = 4

        y = x.check_exp_2d_2m3_nt(arr_test, 4, 0)

        self.assertEqual(y.args["success"], True)

        arr_test[0, 3] = 5

        np.testing.assert_array_equal(y.args["arr"], arr_test)

    def test_logical_param_array(self):
        assert np.all(
            x.const_logical_arr
            == np.array([True, False, True, False, True], dtype=bool)
        )



# SPDX-License-Identifier: GPL-2.0+
import ctypes
import os
import select
import collections
import functools
from dataclasses import dataclass

from .fVar import fVar
from .fVar_t import fVar_t

_TEST_FLAG = os.environ.get("_GFORT2PY_TEST_FLAG") is not None


@dataclass
class variable:
    value: "typing.Any"
    fvar: "typing.Any"
    symbol_ref: int


class _captureStdOut:
    def read_pipe(self, pipe_out):
        def more_data():
            r, _, _ = select.select([pipe_out], [], [], 0)
            return bool(r)

        out = b""
        while more_data():
            out += os.read(pipe_out, 1024)
        return out.decode()

    def __enter__(self):
        if _TEST_FLAG:
            self.pipe_out, self.pipe_in = os.pipe()
            self.stdout = os.dup(1)
            os.dup2(self.pipe_in, 1)

    def __exit__(self, *args, **kwargs):
        if _TEST_FLAG:
            os.dup2(self.stdout, 1)
            print(self.read_pipe(self.pipe_out))
            os.close(self.pipe_in)
            os.close(self.pipe_out)
            os.close(self.stdout)


class fProc:
    Result = collections.namedtuple("Result", ["result", "args"])

    def __init__(self, lib, obj, allobjs, **kwargs):
        self._allobjs = allobjs
        self.obj = obj
        self._lib = lib
        self._return_value = None

        self._func = getattr(lib, self.mangled_name)

    @property
    def mangled_name(self):
        return self.obj.mangled_name

    def in_dll(self, lib):
        return self._func

    @property
    def module(self):
        return self.obj.head.module

    @property
    def name(self):
        return self.obj.name

    def __call__(self, *args, **kwargs):
        self._set_return()

        func_args = self._convert_args(*args, **kwargs)

        with _captureStdOut() as cs:
            if func_args is not None:
                res = self._func(*func_args)
            else:
                res = self._func()

        return self._convert_result(res, func_args)

    def _set_return(self):
        if self.obj.is_subroutine():
            self._func.restype = None  # Subroutine
        else:
            if (
                self.return_var.obj.is_char()
            ):  # Returning a character is done as a character + len at start of arg list
                self._func.restype = None
            else:
                self._func.restype = self.return_var.ctype()

    def args_start(self):
        res = []
        if self.obj.is_function():
            if self.return_var.obj.is_char():
                l = self.return_var.len()
                res.append(self.return_var.from_param(" " * l))
                res.append(self.return_var.ctype_len())

        return res

    def args_check(self, *args, **kwargs):
        count = 0
        arguments = []
        # Build list of inputs
        for fval in self.obj.args():
            var = fVar(self._allobjs[fval.ref], allobjs=self._allobjs)

            try:
                x = kwargs[var.name]
            except KeyError:
                if count <= len(args):
                    x = args[count]
                    count = count + 1
                else:
                    raise TypeError("Not enough arguments passed")

            if x is None and not var.obj.is_optional() and not var.obj.is_dummy():
                raise ValueError(f"Got None for {var.name}")

            if isinstance(x, fVar_t):
                var = x
                x = var.value

            arguments.append(variable(x, var, fval.ref))

        return arguments

    def args_convert(self, input_args):
        args = []
        args_end = []
        # Convert to ctypes
        for var in input_args:
            _, a, e = var.fvar.to_proc(var.value, input_args)
            args.append(a)
            if e is not None:
                args_end.append(e)

        return args, args_end

    def _convert_args(self, *args, **kwargs):
        args_start = self.args_start()

        self.input_args = self.args_check(*args, **kwargs)

        args_mid, args_end = self.args_convert(self.input_args)

        return args_start + args_mid + args_end

    def _convert_result(self, result, args):
        res = {}

        if self.obj.is_function():
            if self.return_var.obj.is_char():
                result = args[0]
                _ = args.pop(0)
                _ = args.pop(0)  # Twice to pop first and second value

        if len(self.obj.args()):
            for var in self.input_args:
                if var.fvar.unpack:
                    try:
                        x = ptr_unpack(var.fvar.value)
                    except AttributeError:  # unset optional arguments
                        x = None
                else:
                    x = var.fvar.cvalue

                name = self._allobjs[var.symbol_ref].name
                if hasattr(x, "_type_"):
                    res[name] = var.fvar.from_ctype(x)
                else:
                    res[name] = x

        if self.obj.is_function():
            result = self.return_var.from_ctype(result)

        return self.Result(result, res)

    def __repr__(self):
        return self.__doc__

    @property
    def __doc__(self):
        if self.obj.is_subroutine():
            ftype = f"subroutine {self.name}"
        else:
            ftype = f"{self.return_var.__doc__()} function {self.name}"

        args = []
        for fval in self.obj.args():
            args.append(fVar(fval.ref, allobjs=self._allobjs).__doc__)

        args = ", ".join(args)
        return f"{ftype} ({args})"

    @property
    def return_var(self):
        if self._return_value is None:
            self._return_value = fVar(
                self._allobjs[self.obj.return_arg()], allobjs=self._allobjs
            )
        return self._return_value


def ptr_unpack(ptr):
    x = ptr
    if hasattr(ptr, "contents"):
        if hasattr(ptr.contents, "contents"):
            x = ptr.contents.contents
        else:
            x = ptr.contents
    return x



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/strings.{gf.lib_ext()}"
MOD = "./tests/strings.mod"

x = gf.fFort(SO, MOD)


class TestStringMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_a_str(self):
        v = "123456798 "
        x.a_str = v
        self.assertEqual(x.a_str, v)

    def test_a_str_bad_length(self):
        v = "132456789kjhgjhf"
        x.a_str = v
        self.assertEqual(x.a_str, v[0:10])

    def test_sub_str_in_explicit(self, capfd):
        v = "1324567980"

        y = x.sub_str_in_explicit(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), v)

    def test_sub_str_in_implicit(self, capfd):
        v = "123456789"

        y = x.sub_str_in_implicit(v)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), v)

    def test_sub_str_multi(self, capfd):
        v = 5
        u = "123456789"
        w = 4

        y = x.sub_str_multi(v, u, w)
        out, err = capfd.readouterr()
        self.assertEqual(out.strip(), str(v + w) + " " + u)

    def test_sub_str_p(self, capfd):
        y = x.sub_str_p("abcdef")
        out, err = capfd.readouterr()
        assert err == ""
        self.assertEqual(y.args["zzz"], "xyzxyz")
        self.assertEqual(out.strip(), "abcdef")

    def test_func_ret_str(self):
        y = x.func_ret_str("abcde")
        self.assertEqual(y.result, "Abcde")

    @pytest.mark.skip("Skipping")
    # We need to call a func on the argument before passing it to func_str_int_len
    def test_func_str_int_len(self):
        y = x.func_str_int_len(10)

        self.assertEqual(out, "10")

    def test_str_alloc(self):
        self.assertEqual(x.str_alloc, None)  # Empty at start

        x.str_alloc = "abcdefghijklmnop"
        self.assertEqual(x.str_alloc, "abcdefghijklmnop")
        y = x.check_str_alloc(1)
        self.assertEqual(y.result, True)

        x.str_alloc = "12345678        "  # Need to empty the space afterwards
        self.assertEqual(x.str_alloc, "12345678        ")
        y = x.check_str_alloc(2)
        self.assertEqual(y.result, False)

    def test_str_alloc_sub(self):
        z = None
        y = x.sub_str_alloc(z)
        self.assertEqual(y.args["x_alloc"], "abcdef")

        y2 = x.sub_str_alloc2(None)
        self.assertEqual(y2.args["x"], "zxcvbnm")

    @pytest.mark.skip
    def test_str_alloc_sub_realloc(self):
        y2 = x.sub_str_alloc2("qwerty")
        self.assertEqual(y2.args["x"], "asdfghjkl")

    def test_str_array_type_chceck(self):
        with pytest.raises(TypeError) as cm:
            x.a_str_exp_1d = np.zeros(5, dtype=np.unicode_)

    def test_str_func_inout_str(self, capfd):
        z = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
            dtype="S10",
        )

        res = x.str_array_inout(z)
        out, err = capfd.readouterr()

        assert out == " aaaaaaaaaabbbbbbbbbbccccccccccddddddddddeeeeeeeeee\n\n"

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(res.args["x"] == z2)

    def test_func_str_array_dt(self):
        data = {
            "start_guard": 123456789,
            "a_str_exp_1d": np.array(
                ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
                dtype="S10",
            ),
            "end_guard": 123456789,
        }

        assert not x.func_str_array_dt(data).result

        data = {
            "start_guard": 123456789,
            "a_str_exp_1d": np.array(
                ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
                dtype="S10",
            ),
            "end_guard": 123456789,
        }

        assert x.func_str_array_dt(data).result

    def test_func_str_array_dt_alloc(self):
        data = {
            "start_guard": 123456789,
            "b_str_alloc_1d": None,
            "end_guard": 123456789,
        }

        res = x.func_str_array_dt_alloc(data)

        assert res.result

        z = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(z == res.args["x"]["b_str_alloc_1d"])

    def test_str_func_inout_str2(self, capfd):
        z = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
            dtype="S10",
        )

        res = x.str_array_inout2(z, 5)
        out, err = capfd.readouterr()

        assert out == " aaaaaaaaaabbbbbbbbbbccccccccccddddddddddeeeeeeeeee\n\n"

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(res.args["x"] == z2)

    def test_str_func_inout_str3(self, capfd):
        z = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
            dtype="S10",
        )

        res = x.str_array_inout3(z)
        out, err = capfd.readouterr()

        assert out == " aaaaaaaaaabbbbbbbbbbccccccccccddddddddddeeeeeeeeee\n\n"

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(res.args["x"] == z2)

    def test_str_func_inout_alloc(self, capfd):
        z = None

        res = x.str_array_allocate(z)

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(res.args["x"] == z2)

    @pytest.mark.skip
    def test_str_func_inout_str4(self, capfd):
        z = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
            dtype="S10",
        )

        res = x.str_array_inout4(z)
        out, err = capfd.readouterr()

        assert out == " aaaaaaaaaabbbbbbbbbbccccccccccddddddddddeeeeeeeeee\n\n"

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(res.args["x"] == z2)

    def test_str_array_param(self):
        assert np.all(x.a_str_p_1d == np.array(["aa", "bb", "cc"], dtype="S2"))

    def test_check_a_str_exp_1d(self, capfd):
        x.a_str_exp_1d = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb", "cccccccccc", "dddddddddd", "eeeeeeeeee"],
            dtype="S10",
        )

        res = x.check_a_str_exp_1d()
        out, err = capfd.readouterr()

        assert out == " aaaaaaaaaabbbbbbbbbbccccccccccddddddddddeeeeeeeeee\n\n"

        z2 = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        assert np.all(x.a_str_exp_1d == z2)

    def test_b_str_alloc_1d(self):
        x.b_str_alloc_1d = np.array(
            ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
            dtype="S10",
        )

        res = x.check_b_str_alloc_1d()

        assert res.result

        x.b_str_alloc_1d = np.array(
            ["aaaaaaaaaa", "bbbbbbbbbb"],
            dtype="S10",
        )

        res = x.check_b_str_alloc_1d()

        assert not res.result

    def test_alloc_b_str_alloc_1d(self):
        res = x.alloc_b_str_alloc_1d()

        assert np.all(
            x.b_str_alloc_1d
            == np.array(
                ["zzzzzzzzzz", "yyyyyyyyyy", "qqqqqqqqqq", "wwwwwwwwww", "xxxxxxxxxx"],
                dtype="S10",
            )
        )

    def test_set_str_array_dt_out(self):
        res = x.set_str_array_dt_out({})

        assert res.args["x"]["a_str1"] == "qwertyuiop[]"
        assert res.args["x"]["a_str2"] == "asdfghjkl;zx"
        assert res.args["x"]["a_int"] == 99



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/pointers.{gf.lib_ext()}"
MOD = "./tests/ptrs.mod"

x = gf.fFort(SO, MOD)


class TestPtrsMethods:
    def assertEqual(self, x, y):
        assert x == y

    def test_a_int_point(self):
        v = 1
        x.a_int_point = v
        self.assertEqual(x.a_int_point, v)

    def test_a_int_lp_point(self):
        v = 1
        x.a_int_lp_point = v
        self.assertEqual(x.a_int_lp_point, v)

    def test_a_real_point(self):
        v = 1.0
        x.a_real_point = v
        self.assertEqual(x.a_real_point, v)

    def test_a_real_dp_point(self):
        v = 1.0
        x.a_real_dp_point = v
        self.assertEqual(x.a_real_dp_point, v)

    def test_a_real_qp_point(self):
        v = 1.0
        with pytest.raises(NotImplementedError) as cm:
            x.a_real_qp_point = v
            self.assertEqual(x.a_real_qp_point, v)

    def test_a_str_point(self):
        v = "abcdefghij"
        x.a_str_point = v
        self.assertEqual(x.a_str_point, v)

    def test_a_int_target(self):
        v = 1
        x.a_int_target = v
        self.assertEqual(x.a_int_target, v)

    def test_a_int_lp_target(self):
        v = 1
        x.a_int_lp_target = v
        self.assertEqual(x.a_int_lp_target, v)

    def test_a_real_target(self):
        v = 1.0
        x.a_real_target = v
        self.assertEqual(x.a_real_target, v)

    def test_a_real_dp_target(self):
        v = 1.0
        x.a_real_dp_target = v
        self.assertEqual(x.a_real_dp_target, v)

    def test_a_real_qp_target(self):
        v = 1.0
        with pytest.raises(NotImplementedError) as cm:
            x.a_real_qp_target = v
            self.assertEqual(x.a_real_qp_target, v)

    def test_a_str_target(self):
        v = "abcdefghij"
        x.a_str_target = v
        self.assertEqual(x.a_str_target, v)



# SPDX-License-Identifier: GPL-2.0+

import operator

default_ops = {
    "UPLUS": operator.__add__,
    "UMINUS": operator.__sub__,
    "PLUS": operator.__add__,
    "MINUS": operator.__sub__,
    "TIMES": operator.__mul__,
    "DIVIDE": operator.__truediv__,
    "POWER": operator.__pow__,
    "CONCAT": operator.__add__,  # Only for strings
    "AND": operator.__and__,
    "OR": operator.__or__,
    "EQV": operator.__eq__,
    "NEQV": operator.__ne__,
    "EQ_SIGN": operator.__eq__,
    "EQ": operator.__eq__,
    "NE_SIGN": operator.__ne__,
    "NE": operator.__ne__,
    "GT_SIGN": operator.__gt__,
    "GT": operator.__gt__,
    "GE_SIGN": operator.__ge__,
    "GE": operator.__ge__,
    "LT_SIGN": operator.__le__,
    "LT": operator.__le__,
    "LE_SIGN": operator.__le__,
    "LE": operator.__le__,
    "NOT": operator.__not__,
    "PARENTHESES": None,
    "USER": None,
    "NULL": None,
}


def run_unary(unary, x, y, *, ops=default_ops):
    if unary == "PARENTHESES":
        return x

    op = ops[unary]

    if op is None:
        raise NotImplementedError(f"Unary op {unary} not implemented")

    return op(x, y)
